"""Tests for DAG construction, orchestration, and lookup behaviour in the client."""

import json
from unittest.mock import Mock, patch

import polars as pl
import pyarrow as pa
import pytest
from httpx import Response
from polars.testing import assert_frame_equal
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG, DAGNodeExecutionStatus
from matchbox.client.models import Model
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.client.resolvers import Components, Resolver
from matchbox.client.sources import Source
from matchbox.common.arrow import SCHEMA_QUERY_WITH_LEAVES, table_to_buffer
from matchbox.common.dtos import (
    Collection,
    CRUDOperation,
    DefaultGroup,
    ErrorResponse,
    Match,
    PermissionGrant,
    PermissionType,
    ResourceOperationStatus,
    Run,
    Step,
    StepName,
    StepPath,
    StepType,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxEmptyServerResponse,
    MatchboxStepNotFoundError,
    MatchboxStepTypeError,
)
from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.resolvers import resolver_factory
from matchbox.common.factories.sources import (
    linked_sources_factory,
    source_factory,
)


def test_dag_list(matchbox_api: MockRouter) -> None:
    """Can retrieve list of DAG names."""
    dummy_names = ["companies", "contacts"]
    matchbox_api.get("/collections").mock(return_value=Response(200, json=dummy_names))
    assert DAG.list_all() == ["companies", "contacts"]


@patch.object(Source, "run")
@patch.object(Model, "run")
@patch.object(Resolver, "run")
@patch.object(Source, "sync")
@patch.object(Model, "sync")
@patch.object(Resolver, "sync")
@patch.object(Source, "clear_data")
@patch.object(Model, "clear_data")
@patch.object(Resolver, "clear_data")
def test_dag_run_and_sync(
    resolver_clear_mock: Mock,
    model_clear_mock: Mock,
    source_clear_mock: Mock,
    resolver_sync_mock: Mock,
    model_sync_mock: Mock,
    source_sync_mock: Mock,
    resolver_run_mock: Mock,
    model_run_mock: Mock,
    source_run_mock: Mock,
    sqla_sqlite_warehouse: Engine,
) -> None:
    """A legal DAG can be built and run."""
    # Set up constituents
    foo_tkit = source_factory(
        name="foo", engine=sqla_sqlite_warehouse
    ).write_to_location()
    bar_tkit = source_factory(
        name="bar", engine=sqla_sqlite_warehouse
    ).write_to_location()
    baz_tkit = source_factory(
        name="baz", engine=sqla_sqlite_warehouse
    ).write_to_location()

    dag = TestkitDAG().dag

    # Structure: sources can be added
    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    baz = dag.source(**baz_tkit.into_dag())

    assert set(dag.nodes.keys()) == {foo.name, bar.name, baz.name}

    # Structure: sources can be deduped
    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )

    # Structure:
    # - sources can be passed directly to linkers
    # - or, linkers can take dedupers
    foo_bar = foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    # Additional independent model
    foo_baz = foo.query().linker(
        baz.query(),
        name="foo_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    root = foo_bar.resolver(d_foo, foo_baz, name="root", resolver_class=Components)

    assert set(dag.nodes.keys()) == {
        foo.name,
        bar.name,
        baz.name,
        d_foo.name,
        foo_bar.name,
        foo_baz.name,
        root.name,
    }

    # Run DAG
    dag.run_and_sync()

    assert source_run_mock.call_count == 3
    assert source_sync_mock.call_count == 3
    assert model_run_mock.call_count == 3
    assert model_sync_mock.call_count == 3
    assert resolver_run_mock.call_count == 1
    assert resolver_sync_mock.call_count == 1
    source_clear_mock.assert_not_called()
    model_clear_mock.assert_not_called()
    resolver_clear_mock.assert_not_called()

    # Running DAG destroys intermediate results
    dag.run_and_sync(low_memory=True)
    assert source_clear_mock.call_count == 3
    assert model_clear_mock.call_count == 3
    assert resolver_clear_mock.call_count == 1


@patch.object(Source, "sync")
@patch.object(Model, "sync")
@patch.object(Resolver, "sync")
@patch.object(Source, "run", autospec=True)
@patch.object(Model, "run", autospec=True)
@patch.object(Resolver, "run", autospec=True)
def test_dag_run_and_sync_orders_shared_parents_before_children(
    resolver_run_mock: Mock,
    model_run_mock: Mock,
    source_run_mock: Mock,
    resolver_sync_mock: Mock,
    model_sync_mock: Mock,
    source_sync_mock: Mock,
    sqla_sqlite_warehouse: Engine,
) -> None:
    """Execution order ensures all structural parents run before their children."""
    foo_tkit = source_factory(
        name="foo", engine=sqla_sqlite_warehouse
    ).write_to_location()
    bar_tkit = source_factory(
        name="bar", engine=sqla_sqlite_warehouse
    ).write_to_location()
    baz_tkit = source_factory(
        name="baz", engine=sqla_sqlite_warehouse
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    baz = dag.source(**baz_tkit.into_dag())

    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    foo_bar = foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    foo_baz = foo.query().linker(
        baz.query(),
        name="foo_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    _ = foo_bar.resolver(
        d_foo,
        foo_baz,
        name="root",
        resolver_class=Components,
        resolver_settings={
            "thresholds": {d_foo.name: 0, foo_bar.name: 0, foo_baz.name: 0}
        },
    )

    run_order: list[str] = []

    def _record_run(
        node: Source | Model | Resolver, *args: object, **kwargs: object
    ) -> pl.DataFrame:
        run_order.append(node.name)
        node._local_data = pl.DataFrame()
        return node._local_data

    source_run_mock.side_effect = _record_run
    model_run_mock.side_effect = _record_run
    resolver_run_mock.side_effect = _record_run

    dag.run_and_sync()

    assert len(run_order) == len(dag.nodes)
    assert set(run_order) == set(dag.nodes.keys())

    run_index = {name: index for index, name in enumerate(run_order)}
    for child, parents in dag.graph.items():
        for parent in parents:
            assert run_index[parent] < run_index[child]


def test_dags_missing_dependency(sqla_sqlite_warehouse: Engine) -> None:
    """Steps cannot be added before their dependencies."""
    dag = TestkitDAG().dag

    foo = source_factory(name="foo", engine=sqla_sqlite_warehouse, dag=dag).source

    with pytest.raises(ValueError, match="not added to DAG"):
        dag.model(
            left_query=foo.query(),
            name="d_foo",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Failure leads to no dags being added
    assert not len(dag.nodes)
    assert not len(dag.graph)

    d_foo = model_factory(name="d_foo", dag=dag).model

    with pytest.raises(ValueError, match="not added to DAG"):
        dag.resolver(
            inputs=[d_foo],
            name="r_foo",
            resolver_class=Components,
            resolver_settings={},
        )

    # Failure leads to no dags being added
    assert not len(dag.nodes)
    assert not len(dag.graph)


def test_mixing_dags_fails(sqla_sqlite_warehouse: Engine) -> None:
    """Cannot reference a different DAG when adding a step."""
    dag = TestkitDAG().dag
    dag2 = TestkitDAG().dag

    foo_tkit = source_factory(name="foo", engine=sqla_sqlite_warehouse)
    foo = dag.source(**foo_tkit.into_dag())

    with pytest.raises(ValueError, match="mix DAGs"):
        # Different DAG in input
        dag2.model(
            left_query=foo.query(),
            name="d_foo",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Failure leads to no dags being added
    assert not len(dag2.nodes)
    assert not len(dag2.graph)


def test_dag_name_clash(sqla_sqlite_warehouse: Engine) -> None:
    """Under some conditions, nodes can be overwritten."""
    dag = TestkitDAG().dag

    foo_tkit = source_factory(name="foo", engine=sqla_sqlite_warehouse)
    bar_tkit = source_factory(name="bar", engine=sqla_sqlite_warehouse)

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())

    d_foo = foo.query().deduper(
        name="d_foo",
        model_class=NaiveDeduper,
        model_settings={"unique_fields": [foo.f(foo.index_fields[0].name)]},
    )

    # Can re-define nodes, in whatever order
    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    updated_foo_args = foo_tkit.into_dag()
    updated_foo_args["description"] = "new description"
    foo = dag.source(**updated_foo_args)

    assert dag.get_model("d_foo").model_settings.unique_fields == []
    assert dag.get_source(foo_tkit.name).description == "new description"

    # Cannot overwrite source with model
    with pytest.raises(ValueError, match="Cannot re-assign"):
        foo.query().deduper(
            name=foo_tkit.name,
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Cannot overwrite model with source
    updated_foo_args = foo_tkit.into_dag()
    updated_foo_args["name"] = "d_foo"
    with pytest.raises(ValueError, match="Cannot re-assign"):
        dag.source(**updated_foo_args)

    # Cannot change inputs of model
    linker = foo.query().linker(
        bar.query(),
        name="linker",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )

    baz_tkit = source_factory(name="baz", engine=sqla_sqlite_warehouse)
    baz = dag.source(**baz_tkit.into_dag())
    with pytest.raises(ValueError, match="Cannot re-assign"):
        foo.query().linker(
            # Linking different source using linker whose name is taken
            baz.query(),
            name="linker",
            model_class=DeterministicLinker,
            model_settings={"comparisons": "l.field=r.field"},
        )

    # Create a valid resolver, then verify clashes
    root = linker.resolver(d_foo, name="root", resolver_class=Components)

    # Cannot overwrite resolver with source
    updated_root_args = foo_tkit.into_dag()
    updated_root_args["name"] = "root"
    with pytest.raises(ValueError, match="Cannot re-assign"):
        dag.source(**updated_root_args)

    # Cannot overwrite resolver with model
    with pytest.raises(ValueError, match="Cannot re-assign"):
        foo.query().deduper(
            name="root",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": []},
        )

    # Cannot change inputs of resolver
    with pytest.raises(ValueError, match="Cannot re-assign"):
        linker.resolver(name="root", resolver_class=Components)

    # After failed attempts, DAG is as we expect
    assert dag.get_source(foo.name) == foo
    assert dag.get_source(bar.name) == bar
    assert dag.get_source(baz.name) == baz
    assert dag.get_model(d_foo.name) == d_foo
    assert dag.get_model(linker.name) == linker
    assert dag.get_resolver(root.name) == root

    assert dag.graph[foo.name] == []
    assert dag.graph[bar.name] == []
    assert dag.graph[baz.name] == []
    assert dag.graph[d_foo.name] == [foo.name]
    assert dag.graph[linker.name] == [foo.name, bar.name]
    assert set(dag.graph[root.name]) == {linker.name, d_foo.name}


def test_dag_default_resolvers(sqla_sqlite_warehouse: Engine) -> None:
    """Test final_steps property returns all apex nodes."""
    dag = TestkitDAG().dag

    # Empty DAG has no final steps
    assert dag.final_steps == []

    # Single apex
    foo_tkit = source_factory(name="foo", engine=sqla_sqlite_warehouse)
    foo = dag.source(**foo_tkit.into_dag())
    assert dag.final_steps == [foo]

    # Multiple apexes (disconnected)
    bar_tkit = source_factory(name="bar", engine=sqla_sqlite_warehouse)
    bar = dag.source(**bar_tkit.into_dag())
    assert set(dag.final_steps) == {foo, bar}

    # After linking, single apex again
    foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    assert len(dag.final_steps) == 1
    assert dag.final_steps[0].name == "foo_bar"


def test_dag_draw_tree(sqla_sqlite_warehouse: Engine) -> None:
    """Test tree mode rendering of the DAG."""
    foo_tkit = source_factory(
        name="foo", engine=sqla_sqlite_warehouse
    ).write_to_location()
    bar_tkit = source_factory(
        name="bar", engine=sqla_sqlite_warehouse
    ).write_to_location()
    baz_tkit = source_factory(
        name="baz", engine=sqla_sqlite_warehouse
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    baz = dag.source(**baz_tkit.into_dag())
    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    foo_bar = foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    foo_baz = foo.query().linker(
        baz.query(),
        name="foo_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    root = foo_bar.resolver(d_foo, foo_baz, name="root", resolver_class=Components)

    node_names = [
        foo.name,
        bar.name,
        baz.name,
        d_foo.name,
        foo_bar.name,
        foo_baz.name,
        root.name,
    ]

    tree_str = dag.draw()
    lines = tree_str.strip().split("\n")
    head_lines, tree_lines = lines[:3], lines[3:]

    # Header
    assert "Collection" in head_lines[0]
    assert "Run" in head_lines[1]

    # Root node is first
    assert root.name in tree_lines[0]

    # All nodes present
    for node in node_names:
        assert any(node in line for line in tree_lines)

    # Type indicators
    assert "💎" in tree_str
    assert "⚙️" in tree_str
    assert "📄" in tree_str

    # Step number badges
    assert "[1]" in tree_str

    # Shared parents shown in each branch, not globally suppressed
    shared_foo_count = sum(1 for line in tree_lines if f" {foo.name} [" in line)
    assert shared_foo_count == 3

    # Tree formatting characters
    assert any(char in tree_str for char in ["└──", "├──", "│"])

    # __str__
    assert str(dag) == dag.draw()

    # Empty DAG
    assert TestkitDAG().dag.draw() == "Empty DAG"

    # Disconnected components
    disconnected_dag = TestkitDAG().dag
    _ = disconnected_dag.source(
        **source_factory(name="qux", engine=sqla_sqlite_warehouse)
        .write_to_location()
        .into_dag()
    )
    _ = disconnected_dag.source(
        **source_factory(name="quux", engine=sqla_sqlite_warehouse)
        .write_to_location()
        .into_dag()
    )
    disconnected_str = disconnected_dag.draw()
    head_lines, tree_lines = disconnected_str[:3], disconnected_str[3:]
    assert "qux" in tree_lines
    assert "quux" in tree_lines
    assert "\n\n" in tree_lines


def test_dag_draw_list(sqla_sqlite_warehouse: Engine) -> None:
    """Test list mode rendering of the DAG."""
    foo_tkit = source_factory(
        name="foo", engine=sqla_sqlite_warehouse
    ).write_to_location()
    bar_tkit = source_factory(
        name="bar", engine=sqla_sqlite_warehouse
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    foo_bar = foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    root = foo_bar.resolver(name="root", resolver_class=Components)

    list_str = dag.draw(mode="list")
    lines = list_str.strip().split("\n")
    head_lines, list_lines = lines[:3], lines[3:]

    # Same header as tree mode
    assert "Collection" in head_lines[0]
    assert "Run" in head_lines[1]

    # All nodes present
    for node in [foo.name, bar.name, foo_bar.name, root.name]:
        assert any(node in line for line in list_lines)

    # Step numbers prefix each line
    for i, line in enumerate(list_lines, start=1):
        assert line.startswith(f"{i}.")

    # Sources appear before linker, linker before resolver
    foo_bar_idx = next(
        i for i, line in enumerate(list_lines) if line.endswith(foo_bar.name)
    )
    foo_idx = next(i for i, line in enumerate(list_lines) if line.endswith(foo.name))
    bar_idx = next(i for i, line in enumerate(list_lines) if line.endswith(bar.name))
    root_idx = next(i for i, line in enumerate(list_lines) if line.endswith(root.name))
    assert foo_bar_idx > foo_idx
    assert foo_bar_idx > bar_idx
    assert root_idx > foo_bar_idx

    # No tree formatting characters in content lines
    assert not any(char in line for line in list_lines for char in ["└──", "├──", "│"])


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("tree", id="tree"),
        pytest.param("list", id="list"),
    ],
)
def test_dag_draw_status(sqla_sqlite_warehouse: Engine, mode: str) -> None:
    """Test that status indicators render correctly in both draw modes."""
    foo_tkit = source_factory(
        name="foo", engine=sqla_sqlite_warehouse
    ).write_to_location()
    bar_tkit = source_factory(
        name="bar", engine=sqla_sqlite_warehouse
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_tkit.into_dag())
    bar = dag.source(**bar_tkit.into_dag())
    d_foo = foo.query().deduper(
        name="d_foo", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    foo_bar = foo.query().linker(
        bar.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    root = foo_bar.resolver(d_foo, name="root", resolver_class=Components)

    draw_str = dag.draw(
        mode=mode,
        status={
            foo_bar.name: DAGNodeExecutionStatus.DOING,
            d_foo.name: DAGNodeExecutionStatus.DONE,
            foo.name: DAGNodeExecutionStatus.SKIPPED,
            root.name: DAGNodeExecutionStatus.DONE,
        },
    )
    content_lines = draw_str.strip().split("\n")[3:]

    # All four indicators present (bar is awaiting)
    assert "✅" in draw_str
    assert "🔄" in draw_str
    assert "⏸️" in draw_str
    assert "⏭️" in draw_str

    # Indicators are directly adjacent to type indicators
    assert "✅⚙️" in draw_str
    assert "🔄⚙️" in draw_str
    assert "⏸️📄" in draw_str
    assert "⏭️📄" in draw_str
    assert "✅💎" in draw_str

    # Correct indicator per node
    for line in content_lines:
        if foo_bar.name in line:
            assert "🔄" in line
        elif d_foo.name in line:
            assert "✅" in line
        elif f" {foo.name}" in line:
            assert "⏭️" in line
        elif bar.name in line:
            assert "⏸️" in line
        elif root.name in line:
            assert "✅" in line


# Lookups


def test_get_matches(matchbox_api: MockRouter) -> None:
    """Match lookup can be generated from DAG."""
    # Make dummy data
    foo = source_factory(name="foo", location_name="sqlite")
    bar = source_factory(name="bar", location_name="postgres")
    baz = source_factory(name="baz", location_name="postgres")

    # Nothing is merged - that's OK for this test
    foo_data = pa.Table.from_pylist(
        [
            {"id": 1, "leaf_id": 1, "key": "1"},
            {"id": 2, "leaf_id": 2, "key": "2"},
        ],
        schema=SCHEMA_QUERY_WITH_LEAVES,
    )

    bar_data = pa.Table.from_pylist(
        [
            {"id": 3, "leaf_id": 3, "key": "a"},
            {"id": 4, "leaf_id": 4, "key": "b"},
        ],
        schema=SCHEMA_QUERY_WITH_LEAVES,
    )

    baz_data = pa.Table.from_pylist(
        [
            {"id": 5, "leaf_id": 5, "key": "x"},
            {"id": 6, "leaf_id": 6, "key": "y"},
        ],
        schema=SCHEMA_QUERY_WITH_LEAVES,
    )

    dag = DAG("companies")

    # Mock API
    # In the beginning, no run
    matchbox_api.get(f"/collections/{dag.name}").mock(
        return_value=Response(
            200,
            json=Collection(name=dag.name, runs=[], default_run=None).model_dump(),
        )
    )

    matchbox_api.post(f"/collections/{dag.name}/runs").mock(
        return_value=Response(
            200,
            json=Run(run_id=1, steps={}).model_dump(),
        )
    )

    # Build dummy DAG
    foo_source = dag.source(**foo.into_dag())
    bar_source = dag.source(**bar.into_dag())
    baz_source = dag.source(**baz.into_dag())
    foo_dedupe = foo_source.query().deduper(
        name="foo_dedupe",
        model_class=NaiveDeduper,
        model_settings={"unique_fields": []},
    )
    foo_bar = foo_source.query().linker(
        bar_source.query(),
        name="foo_bar",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    bar_baz = bar_source.query().linker(
        baz_source.query(),
        name="bar_baz",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    foo_bar_baz = foo_bar.resolver(
        foo_dedupe, bar_baz, name="foo_bar_baz", resolver_class=Components
    )
    dag.new_run()

    # Start with a single resolver apex and exercise implicit apex selection.
    matchbox_api.get(
        "/query",
        params={
            "source": "foo",
            "run_id": 1,
            "collection": dag.name,
            "resolver": foo_bar_baz.name,
            "return_leaf_id": "True",
        },
    ).mock(return_value=Response(200, content=table_to_buffer(foo_data).read()))

    matchbox_api.get(
        "/query",
        params={
            "source": "bar",
            "run_id": 1,
            "collection": dag.name,
            "resolver": foo_bar_baz.name,
            "return_leaf_id": "True",
        },
    ).mock(return_value=Response(200, content=table_to_buffer(bar_data).read()))

    matchbox_api.get(
        "/query",
        params={
            "source": "baz",
            "run_id": 1,
            "collection": dag.name,
            "resolver": foo_bar_baz.name,
            "return_leaf_id": "True",
        },
    ).mock(return_value=Response(200, content=table_to_buffer(baz_data).read()))
    apex_resolved = dag.get_matches()
    assert len(apex_resolved.sources) == 3
    assert {source.name for source in apex_resolved.sources} == {"foo", "bar", "baz"}

    # Add a second resolver after the apex assertion.
    foo_bar_resolver = foo_bar.resolver(
        foo_dedupe, name="foo_bar_resolver", resolver_class=Components
    )

    # Intermediate resolver has a narrower source set (foo + bar only).
    matchbox_api.get(
        "/query",
        params={
            "source": "foo",
            "run_id": 1,
            "collection": dag.name,
            "resolver": foo_bar_resolver.name,
            "return_leaf_id": "True",
        },
    ).mock(return_value=Response(200, content=table_to_buffer(foo_data).read()))
    matchbox_api.get(
        "/query",
        params={
            "source": "bar",
            "run_id": 1,
            "collection": dag.name,
            "resolver": foo_bar_resolver.name,
            "return_leaf_id": "True",
        },
    ).mock(return_value=Response(200, content=table_to_buffer(bar_data).read()))

    # Then the new run
    matchbox_api.get(f"/collections/{dag.name}").mock(
        return_value=Response(
            200,
            json=Collection(name=dag.name, runs=[1], default_run=1).model_dump(),
        )
    )

    matchbox_api.get(f"/collections/{dag.name}/runs/1").mock(
        return_value=Response(
            200,
            json=Run(
                run_id=1,
                steps={
                    foo.name: foo.fake_run().source.to_dto(),
                    bar.name: bar.fake_run().source.to_dto(),
                    baz.name: baz.fake_run().source.to_dto(),
                    foo_dedupe.name: Step(
                        fingerprint=b"mock_dedupe",
                        step_type=StepType.MODEL,
                        config=foo_dedupe.config,
                    ),
                    foo_bar.name: Step(
                        fingerprint=b"mock_model_1",
                        step_type=StepType.MODEL,
                        config=foo_bar.config,
                    ),
                    bar_baz.name: Step(
                        fingerprint=b"mock_model_2",
                        step_type=StepType.MODEL,
                        config=bar_baz.config,
                    ),
                    foo_bar_resolver.name: Step(
                        fingerprint=b"mock_resolver_1",
                        step_type=StepType.RESOLVER,
                        config=foo_bar_resolver.config,
                    ),
                    foo_bar_baz.name: Step(
                        fingerprint=b"mock_resolver_2",
                        step_type=StepType.RESOLVER,
                        config=foo_bar_baz.config,
                    ),
                },
            ).model_dump(),
        )
    )

    # No sources are found
    with pytest.raises(MatchboxStepNotFoundError):
        dag.get_matches(resolver=foo_bar_baz.name, source_filter=["nonexistent"])

    with pytest.raises(MatchboxStepNotFoundError):
        dag.get_matches(resolver=foo_bar_baz.name, location_names=["nonexistent"])

    # With URI filter
    uri_filter_resolved = dag.get_matches(
        resolver=foo_bar_baz.name, location_names=["sqlite"]
    )

    assert len(uri_filter_resolved.sources) == 1
    assert uri_filter_resolved.sources[0].name == "foo"

    # With source filter
    source_filter_resolved = dag.get_matches(
        resolver=foo_bar_baz.name, source_filter=["foo"]
    )

    assert len(source_filter_resolved.sources) == 1
    assert source_filter_resolved.sources[0].name == "foo"

    # Select intermediate resolver
    intermediate_res = dag.get_matches(resolver=foo_bar_resolver.name)
    assert len(intermediate_res.sources) == 2
    assert set([s.name for s in intermediate_res.sources]) == {foo.name, bar.name}

    # With no filter
    full_resolved = dag.get_matches(resolver=foo_bar_baz.name)
    assert len(full_resolved.sources) == 3
    assert len(full_resolved.query_results) == 3
    source_names = [
        full_resolved.sources[0].name,
        full_resolved.sources[1].name,
        full_resolved.sources[2].name,
    ]
    foo_index = source_names.index("foo")
    bar_index = source_names.index("bar")
    baz_index = source_names.index("baz")

    assert_frame_equal(full_resolved.query_results[foo_index], pl.from_arrow(foo_data))
    assert_frame_equal(full_resolved.query_results[bar_index], pl.from_arrow(bar_data))
    assert_frame_equal(full_resolved.query_results[baz_index], pl.from_arrow(baz_data))

    # Retrieve from reconstituted DAG
    loaded_res = DAG("companies").load_default().get_matches(resolver=foo_bar_baz.name)
    assert len(loaded_res.sources) == 3
    assert len(loaded_res.query_results) == 3
    source_names = [
        loaded_res.sources[0].name,
        loaded_res.sources[1].name,
        loaded_res.sources[2].name,
    ]
    foo_index = source_names.index("foo")
    bar_index = source_names.index("bar")
    baz_index = source_names.index("baz")

    assert_frame_equal(loaded_res.query_results[foo_index], pl.from_arrow(foo_data))
    assert_frame_equal(loaded_res.query_results[bar_index], pl.from_arrow(bar_data))
    assert_frame_equal(loaded_res.query_results[baz_index], pl.from_arrow(baz_data))


def test_lookup_key_ok(matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine) -> None:
    """The DAG can map between single keys."""
    # Set up dummy data
    foo_testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="foo"
    ).write_to_location()
    bar_testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="bar"
    ).write_to_location()
    baz_testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="baz"
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_testkit.into_dag())
    bar = dag.source(**bar_testkit.into_dag())
    baz = dag.source(**baz_testkit.into_dag())

    linker_foo_bar = foo.query().linker(
        bar.query(),
        name="linker1",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    linker_bar_baz = bar.query().linker(
        baz.query(),
        name="linker2",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    linker_foo_bar.resolver(linker_bar_baz, name="root", resolver_class=Components)

    foo_path = StepPath(name="foo", collection=dag.name, run=dag.run)
    bar_path = StepPath(name="bar", collection=dag.name, run=dag.run)
    baz_path = StepPath(name="baz", collection=dag.name, run=dag.run)

    mock_match1 = Match(
        cluster=1, source=foo_path, source_id={"a"}, target=bar_path, target_id={"b"}
    )
    mock_match2 = Match(
        cluster=1, source=foo_path, source_id={"a"}, target=baz_path, target_id={"b"}
    )
    # The standard JSON serialiser does not handle Pydantic objects
    serialised_matches = json.dumps(
        [m.model_dump() for m in [mock_match1, mock_match2]]
    )

    get_route = matchbox_api.get("/match").mock(
        return_value=Response(200, content=serialised_matches)
    )

    # Use lookup function
    matches = dag.lookup_key(from_source="foo", to_sources=["bar", "baz"], key="pk1")

    # Verify results
    assert matches == {foo.name: ["a"], bar.name: ["b"], baz.name: ["b"]}
    request = get_route.calls.last.request
    assert request.url.params["collection"] == dag.name
    assert request.url.params["run_id"] == str(dag.run)
    assert request.url.params.get_list("targets") == ["bar", "baz"]
    assert request.url.params["source"] == "foo"
    assert request.url.params["key"] == "pk1"
    assert request.url.params["resolver"] == "root"


def test_resolver_rejects_resolver_inputs(sqla_sqlite_warehouse: Engine) -> None:
    dag = TestkitDAG().dag
    source_testkit = source_factory(
        engine=sqla_sqlite_warehouse,
        name="source",
    ).write_to_location()
    source = dag.source(**source_testkit.into_dag())

    dedupe = source.query().deduper(
        name="source_dedupe",
        model_class=NaiveDeduper,
        model_settings={"unique_fields": []},
    )
    first_resolver = dedupe.resolver(name="resolver_1", resolver_class=Components)

    with pytest.raises(MatchboxStepTypeError, match="Expected one of: model"):
        dedupe.resolver(first_resolver, name="resolver_2", resolver_class=Components)


def test_lookup_key_404_source(matchbox_api: MockRouter) -> None:
    """Key lookup throws a step not found error."""
    # Set up dummy data
    source_testkit = source_factory(name="source")
    target_testkit = source_factory(name="target")

    dag = TestkitDAG().dag

    source = dag.source(**source_testkit.into_dag())
    target = dag.source(**target_testkit.into_dag())
    linker = source.query().linker(
        target.query(),
        name="root",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    source_dedupe = source.query().deduper(
        name="source_dedupe",
        model_class=NaiveDeduper,
        model_settings={"unique_fields": []},
    )
    linker.resolver(source_dedupe, name="root_resolver", resolver_class=Components)

    matchbox_api.get("/match").mock(
        return_value=Response(
            404,
            json=ErrorResponse(
                exception_type="MatchboxStepNotFoundError",
                message="Step 42 not found",
            ).model_dump(),
        )
    )

    # Use match function
    with pytest.raises(MatchboxStepNotFoundError, match="42"):
        dag.lookup_key(from_source="source", to_sources=["target"], key="pk1")


def test_lookup_key_no_matches(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Key lookup raises MatchboxEmptyServerResponse when no matches are found."""
    # Set up dummy data
    source_testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="source"
    ).write_to_location()
    target_testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="target"
    ).write_to_location()

    dag = TestkitDAG().dag

    source = dag.source(**source_testkit.into_dag())
    target = dag.source(**target_testkit.into_dag())
    linker = source.query().linker(
        target.query(),
        name="root",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.field=r.field"},
    )
    source_dedupe = source.query().deduper(
        name="source_dedupe",
        model_class=NaiveDeduper,
        model_settings={"unique_fields": []},
    )
    linker.resolver(source_dedupe, name="root_resolver", resolver_class=Components)

    # Mock empty match results
    matchbox_api.get("/match").mock(return_value=Response(200, content="[]"))

    # Test that empty match results raise MatchboxEmptyServerResponse
    with pytest.raises(
        MatchboxEmptyServerResponse, match="The match operation returned no data"
    ):
        dag.lookup_key(from_source="source", to_sources=["target"], key="pk1")


def test_from_dto() -> None:
    """Test reconstructing Sources, Models and Resolvers from a Step."""
    # Setup
    dag_testkit = TestkitDAG()
    linked_testkit = linked_sources_factory(dag=dag_testkit.dag)
    dag_testkit.add_linked_sources(linked_testkit)
    true_entities = tuple(linked_testkit.true_entities)

    crn_testkit = linked_testkit.sources["crn"].fake_run()
    dh_testkit = linked_testkit.sources["dh"].fake_run()
    cdms_testkit = linked_testkit.sources["cdms"].fake_run()

    crn_dh_model = model_factory(
        name="link_crn_dh",
        left_testkit=crn_testkit,
        right_testkit=dh_testkit,
        true_entities=true_entities,
    ).fake_run()
    crn_cdms_model = model_factory(
        name="link_crn_cdms",
        left_testkit=crn_testkit,
        right_testkit=cdms_testkit,
        true_entities=true_entities,
    ).fake_run()
    dag_testkit.add_model(crn_dh_model)
    dag_testkit.add_model(crn_cdms_model)

    resolver_testkit = resolver_factory(
        dag=dag_testkit.dag,
        inputs=[crn_dh_model, crn_cdms_model],
        true_entities=true_entities,
    ).fake_run()
    dag_testkit.add_resolver(resolver_testkit)

    # Test 1: Add all steps in dependency order
    t1_dag = TestkitDAG().dag
    for testkit in [crn_testkit, dh_testkit, cdms_testkit]:
        t1_dag.add_step(name=testkit.name, step=testkit.source.to_dto())
    for testkit in [crn_dh_model, crn_cdms_model]:
        t1_dag.add_step(name=testkit.name, step=testkit.model.to_dto())
    t1_dag.add_step(
        name=resolver_testkit.name,
        step=resolver_testkit.resolver.to_dto(),
    )

    assert t1_dag.name == dag_testkit.dag.name
    assert t1_dag.run == dag_testkit.dag.run
    for name, step in t1_dag.nodes.items():
        assert step.config == dag_testkit.dag.nodes[name].config
    assert t1_dag.graph == dag_testkit.dag.graph

    # Test 2: Resolver added before its model dependencies raises
    t2_dag = TestkitDAG().dag
    with pytest.raises(ValueError, match="not found in DAG"):
        t2_dag.add_step(
            name=resolver_testkit.name,
            step=resolver_testkit.resolver.to_dto(),
        )

    # Test 3: Model added before its source dependencies raises
    t3_dag = TestkitDAG().dag
    with pytest.raises(ValueError, match="not found in DAG"):
        t3_dag.add_step(
            name=crn_dh_model.name,
            step=crn_dh_model.model.to_dto(),
        )


def test_dag_creates_new_collection(matchbox_api: MockRouter) -> None:
    """Connect creates a new collection when it doesn't exist."""
    dag = DAG(name="test_collection")

    # Mock collection not found, then found after creation
    matchbox_api.get("/collections/test_collection").mock(
        side_effect=[
            Response(
                404,
                json=ErrorResponse(
                    exception_type="MatchboxCollectionNotFoundError",
                    message="Collection not found",
                ).model_dump(),
            ),
            Response(
                200,
                json=Collection(
                    name="test_collection",
                    runs=[],
                    default_run=None,
                ).model_dump(),
            ),
        ]
    )

    # Mock collection creation
    create_route = matchbox_api.post("/collections/test_collection").mock(
        return_value=Response(
            200,
            json=ResourceOperationStatus(
                success=True,
                target="Collection test_collection",
                operation=CRUDOperation.CREATE,
            ).model_dump(),
        )
    )

    # Mock run creation
    matchbox_api.post("/collections/test_collection/runs").mock(
        return_value=Response(
            200,
            json=Run(run_id=1, steps={}).model_dump(),
        )
    )

    # Connect the DAG
    result = dag.new_run()

    # Verify
    assert result == dag
    assert dag.run == 1

    assert create_route.called
    assert json.loads(create_route.calls.last.request.content) == [
        PermissionGrant(
            group_name=DefaultGroup.PUBLIC, permission=PermissionType.ADMIN
        ).model_dump()
    ]


def test_dag_creates_new_collection_with_custom_admin(
    matchbox_api: MockRouter,
) -> None:
    """DAG can be configured with a custom admin group."""
    dag = DAG(name="test_collection", admin_group="custom_admins")

    # Mock collection not found, then found after creation
    matchbox_api.get("/collections/test_collection").mock(
        side_effect=[
            Response(
                404,
                json=ErrorResponse(
                    exception_type="MatchboxCollectionNotFoundError",
                    message="Collection not found",
                ).model_dump(),
            ),
            Response(
                200,
                json=Collection(
                    name="test_collection",
                    runs=[],
                    default_run=None,
                ).model_dump(),
            ),
        ]
    )

    # Mock collection creation
    create_route = matchbox_api.post("/collections/test_collection").mock(
        return_value=Response(
            200,
            json=ResourceOperationStatus(
                success=True,
                target="Collection test_collection",
                operation=CRUDOperation.CREATE,
            ).model_dump(),
        )
    )

    # Mock run creation
    matchbox_api.post("/collections/test_collection/runs").mock(
        return_value=Response(
            200,
            json=Run(run_id=1, steps={}).model_dump(),
        )
    )

    dag.new_run()

    assert create_route.called
    assert json.loads(create_route.calls.last.request.content) == [
        PermissionGrant(
            group_name="custom_admins", permission=PermissionType.ADMIN
        ).model_dump()
    ]


@pytest.mark.parametrize(
    ("has_existing_runs", "expected_run_id"),
    [
        pytest.param(False, 1, id="no_existing_runs"),
        pytest.param(True, 4, id="with_existing_runs"),
    ],
)
def test_dag_uses_existing_collection(
    matchbox_api: MockRouter, has_existing_runs: bool, expected_run_id: int
) -> None:
    """New runs can be started from existing collection."""
    dag = DAG(name="test_collection")

    # Mock existing collection
    existing_runs = [2, 3] if has_existing_runs else []
    matchbox_api.get("/collections/test_collection").mock(
        return_value=Response(
            200,
            json=Collection(
                name="test_collection",
                runs=existing_runs,
                default_run=None,
            ).model_dump(),
        )
    )

    # Mock deleting non-default runs
    if has_existing_runs:
        for run_id in existing_runs:
            matchbox_api.delete(f"/collections/test_collection/runs/{run_id}").mock(
                return_value=Response(
                    200,
                    json=ResourceOperationStatus(
                        success=True,
                        target=f"Run {run_id}",
                        operation=CRUDOperation.DELETE,
                    ).model_dump(),
                )
            )

    # Mock run creation
    matchbox_api.post("/collections/test_collection/runs").mock(
        return_value=Response(
            200,
            json=Run(run_id=expected_run_id, steps={}).model_dump(),
        )
    )

    # Connect the DAG
    result = dag.new_run()

    # Verify
    assert result == dag
    assert dag.run == expected_run_id


def test_dag_load_server_run(matchbox_api: MockRouter) -> None:
    """Can retrieve serialised DAG from the server."""
    # Setup
    dag_testkit = TestkitDAG()
    linked_testkit = linked_sources_factory(dag=dag_testkit.dag)
    dag_testkit.add_linked_sources(linked_testkit)
    true_entities = tuple(linked_testkit.true_entities)

    crn_testkit = linked_testkit.sources["crn"].fake_run()
    dh_testkit = linked_testkit.sources["dh"].fake_run()
    cdms_testkit = linked_testkit.sources["cdms"].fake_run()

    crn_dh_model = model_factory(
        name="link_crn_dh",
        left_testkit=crn_testkit,
        right_testkit=dh_testkit,
        true_entities=true_entities,
    ).fake_run()
    crn_cdms_model = model_factory(
        name="link_crn_cdms",
        left_testkit=crn_testkit,
        right_testkit=cdms_testkit,
        true_entities=true_entities,
    ).fake_run()
    dag_testkit.add_model(crn_dh_model)
    dag_testkit.add_model(crn_cdms_model)

    resolver_testkit = resolver_factory(
        dag=dag_testkit.dag,
        inputs=[crn_dh_model, crn_cdms_model],
        true_entities=true_entities,
    ).fake_run()
    dag_testkit.add_resolver(resolver_testkit)

    # Full pipeline: 3 sources, 2 linkers, 1 resolver
    steps: dict[StepName, Step] = {
        crn_testkit.name: crn_testkit.source.to_dto(),
        dh_testkit.name: dh_testkit.source.to_dto(),
        cdms_testkit.name: cdms_testkit.source.to_dto(),
        crn_dh_model.name: crn_dh_model.model.to_dto(),
        crn_cdms_model.name: crn_cdms_model.model.to_dto(),
        resolver_testkit.name: resolver_testkit.resolver.to_dto(),
    }

    run = Run(run_id=1, steps=steps)

    matchbox_api.get(f"/collections/{dag_testkit.dag.name}").mock(
        return_value=Response(
            200,
            json=Collection(
                name=dag_testkit.dag.name,
                runs=[1, 2],
                default_run=1,
            ).model_dump(),
        )
    )
    matchbox_api.get(f"/collections/{dag_testkit.dag.name}/runs/1").mock(
        return_value=Response(200, json=run.model_dump())
    )

    # Load default run and verify reconstruction matches original DAG
    default_dag = DAG(name=dag_testkit.dag.name)
    default_dag = default_dag.load_default()

    assert default_dag.name == dag_testkit.dag.name
    assert default_dag.run == 1
    assert set(default_dag.nodes.keys()) == set(dag_testkit.dag.nodes.keys())
    assert default_dag.graph == dag_testkit.dag.graph

    matchbox_api.get(f"/collections/{dag_testkit.dag.name}/runs/2").mock(
        return_value=Response(200, json=run.model_dump())
    )

    # Load pending run and verify it gets run=2
    pending_dag = DAG(name=dag_testkit.dag.name)
    pending_dag = pending_dag.load_pending()

    assert pending_dag.name == dag_testkit.dag.name
    assert pending_dag.run == 2
    assert set(pending_dag.nodes.keys()) == set(dag_testkit.dag.nodes.keys())
    assert pending_dag.graph == dag_testkit.dag.graph

    # Compatible local nodes are overwritten by server state
    overwritten_dag = DAG(name=dag_testkit.dag.name)
    overwritten_source = overwritten_dag.source(**crn_testkit.into_dag())
    overwritten_source.description = "new description"
    overwritten_dag.load_pending()
    assert (
        overwritten_dag.get_source(crn_testkit.name).description
        == crn_testkit.source.description
    )

    # A local node with the same name but different parents cannot be reconciled
    # This mirrors a script that defines a same-name node with changed dependencies.
    clashing_dag = DAG(name=dag_testkit.dag.name)
    crn = clashing_dag.source(**crn_testkit.into_dag())
    clashing_dag.source(**dh_testkit.into_dag())
    crn.query().deduper(
        name=crn_dh_model.name,
        model_class=NaiveDeduper,
        model_settings={"unique_fields": []},
    )
    with pytest.raises(ValueError, match="Cannot re-assign"):
        clashing_dag.load_pending()

    # Missing collection raises
    matchbox_api.get(f"/collections/{dag_testkit.dag.name}/runs/1").mock(
        return_value=Response(
            404,
            json=ErrorResponse(
                exception_type="MatchboxCollectionNotFoundError",
                message="Collection not found",
            ).model_dump(),
        ),
    )
    with pytest.raises(MatchboxCollectionNotFoundError):
        DAG(name=dag_testkit.dag.name).load_default()


def test_dag_load_run_complex_dependencies(matchbox_api: MockRouter) -> None:
    """Test that _load_run handles nodes with complex dependencies.

    In particular, tests for DAGs where nodes have the same dependency count but
    inter-dependencies between them.
    """
    test_dag = TestkitDAG().dag

    linked_testkit = linked_sources_factory(dag=test_dag)
    crn_testkit = linked_testkit.sources["crn"].fake_run()
    dh_testkit = linked_testkit.sources["dh"].fake_run()
    cdms_testkit = linked_testkit.sources["cdms"].fake_run()

    # model_inner depends on two sources.
    model_inner_testkit = model_factory(
        name="model_inner",
        left_testkit=crn_testkit,
        right_testkit=dh_testkit,
        true_entities=linked_testkit.true_entities,
        dag=test_dag,
    ).fake_run()

    model_side_testkit = model_factory(
        name="model_side",
        left_testkit=dh_testkit,
        right_testkit=cdms_testkit,
        true_entities=linked_testkit.true_entities,
        dag=test_dag,
    ).fake_run()

    # Add sources and models to the reference DAG.
    test_dag.source(**crn_testkit.into_dag())
    test_dag.source(**dh_testkit.into_dag())
    test_dag.source(**cdms_testkit.into_dag())
    test_dag.model(**model_inner_testkit.into_dag())
    test_dag.model(**model_side_testkit.into_dag())

    resolver_inner_testkit = resolver_factory(
        dag=test_dag,
        name="resolver_inner",
        inputs=[model_inner_testkit, model_side_testkit],
        true_entities=tuple(linked_testkit.true_entities),
    ).fake_run()
    test_dag.add_step(
        name=resolver_inner_testkit.name,
        step=resolver_inner_testkit.resolver.to_dto(),
    )
    resolver_inner = test_dag.get_resolver(resolver_inner_testkit.name)

    # model_outer also has dependency count=2, but one dependency is resolver_inner.
    # This must still load after resolver_inner, even if dict order puts it first.
    model_outer = test_dag.model(
        name="model_outer",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.key=r.key"},
        left_query=resolver_inner.query(),
        right_query=test_dag.get_source(cdms_testkit.name).query(),
    )

    # Intentionally place model_outer first: dependency count ties with model_inner,
    # but model_outer depends on resolver_inner and must therefore be delayed.
    steps: dict[StepName, Step] = {
        model_outer.name: Step(
            fingerprint=b"model_outer",
            step_type=StepType.MODEL,
            config=model_outer.config,
        ),
        crn_testkit.name: crn_testkit.source.to_dto(),
        dh_testkit.name: dh_testkit.source.to_dto(),
        cdms_testkit.name: cdms_testkit.source.to_dto(),
        model_inner_testkit.name: model_inner_testkit.model.to_dto(),
        model_side_testkit.name: model_side_testkit.model.to_dto(),
        resolver_inner_testkit.name: resolver_inner_testkit.resolver.to_dto(),
    }

    run = Run(run_id=1, steps=steps)

    # Mock API
    matchbox_api.get(f"/collections/{test_dag.name}").mock(
        return_value=Response(
            200,
            json=Collection(
                name=test_dag.name,
                runs=[1],
                default_run=1,
            ).model_dump(),
        )
    )

    matchbox_api.get(f"/collections/{test_dag.name}/runs/1").mock(
        return_value=Response(
            200,
            json=run.model_dump(),
        )
    )

    # Load the run
    loaded_dag = DAG(name=test_dag.name)
    loaded_dag = loaded_dag.load_default()

    # Verify all nodes were loaded correctly
    assert loaded_dag.run == 1
    assert set(loaded_dag.nodes.keys()) == set(test_dag.nodes.keys())
    assert loaded_dag.graph == test_dag.graph


def test_dag_set_client(sqla_sqlite_warehouse: Engine) -> None:
    """Client can be set for all sources at once."""
    # Create factory data
    foo_params = source_factory(name="foo").into_dag()
    bar_params = source_factory(name="bar").into_dag()

    # Create new DAG
    dag = DAG(name="dag")
    dag.source(**foo_params)
    dag.source(**bar_params)

    # Setting client re-assigns all clients
    assert dag.get_source("foo").location.client != sqla_sqlite_warehouse
    assert dag.get_source("bar").location.client != sqla_sqlite_warehouse
    dag.set_client(sqla_sqlite_warehouse)
    assert dag.get_source("foo").location.client == sqla_sqlite_warehouse
    assert dag.get_source("bar").location.client == sqla_sqlite_warehouse


def test_dag_set_default_ok(
    matchbox_api: MockRouter,
    sqla_sqlite_warehouse: Engine,
) -> None:
    """Set default makes run immutable and sets as default."""
    # Create test data
    dag = TestkitDAG().dag

    crn = dag.source(**source_factory().into_dag())
    dd = crn.query().deduper(
        name="dd", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    dd.resolver(name="r", resolver_class=Components)

    # Mock set mutable
    api_mutable = matchbox_api.patch(
        f"/collections/{dag.name}/runs/{dag.run}/mutable"
    ).mock(
        return_value=Response(
            200,
            json=ResourceOperationStatus(
                success=True,
                target=f"Run {dag.run}",
                operation=CRUDOperation.UPDATE,
            ).model_dump(),
        )
    )

    # Mock set default
    api_default = matchbox_api.patch(
        f"/collections/{dag.name}/runs/{dag.run}/default"
    ).mock(
        return_value=Response(
            200,
            json=ResourceOperationStatus(
                success=True,
                target=f"Run {dag.run}",
                operation=CRUDOperation.UPDATE,
            ).model_dump(),
        )
    )

    # Set as default
    dag.set_default()

    # Verify both endpoints were called
    assert api_mutable.called
    assert api_default.called


def test_dag_set_default_not_connected() -> None:
    """Set default raises error when DAG is not connected to server."""
    dag = DAG(name="test_collection")
    crn = dag.source(**source_factory().into_dag())
    dd = crn.query().deduper(
        name="dd", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    dd.resolver(name="r", resolver_class=Components)

    with pytest.raises(RuntimeError, match="has not been connected"):
        dag.set_default()


def test_dag_set_default_unreachable_nodes(sqla_sqlite_warehouse: Engine) -> None:
    """Nodes cannot be unreachable from root when setting a default run."""
    dag = TestkitDAG().dag

    foo_tkit = source_factory(name="foo", engine=sqla_sqlite_warehouse)
    bar_tkit = source_factory(name="bar", engine=sqla_sqlite_warehouse)

    foo = dag.source(**foo_tkit.into_dag())
    _ = dag.source(**bar_tkit.into_dag())

    dd = foo.query().deduper(
        name="dd", model_class=NaiveDeduper, model_settings={"unique_fields": []}
    )
    dd.resolver(name="r", resolver_class=Components)

    with pytest.raises(ValueError, match="unreachable"):
        dag.set_default()



"""Authentication routes for the Matchbox server."""

from fastapi import APIRouter, Security

from matchbox.common.dtos import AuthStatusResponse, DefaultUser, LoginResponse
from matchbox.server.api.dependencies import (
    JWT_HEADER,
    BackendDependency,
    CurrentUserDependency,
    SettingsDependency,
    validate_jwt,
)

router = APIRouter(prefix="/auth", tags=["authentication"])


@router.post("/login")
def login(
    backend: BackendDependency,
    settings: SettingsDependency,
    client_token: str = Security(JWT_HEADER),
) -> LoginResponse:
    """Upsert the user found in the JWT.

    If in setup mode, will add the user to the admins group.
    """
    user = validate_jwt(settings=settings, client_token=client_token)
    return backend.login(user=user)


@router.get("/status")
async def authentication_status(
    settings: SettingsDependency,
    user: CurrentUserDependency,
) -> AuthStatusResponse:
    """Check authentication status and return user details.

    Returns the current authentication status and user.
    """
    if not settings.authorisation:
        return AuthStatusResponse(authenticated=False)

    is_authenticated = user is not None and user.user_name != DefaultUser.PUBLIC

    return AuthStatusResponse(authenticated=is_authenticated, user=user)



"""Factory helpers for resolver testkits."""

from collections.abc import Iterable, Mapping
from typing import Annotated, Any, ClassVar, Self

import polars as pl
from faker import Faker
from pydantic import BaseModel, ConfigDict, Field

from matchbox.client.dags import DAG
from matchbox.client.models import Model
from matchbox.client.queries import Query
from matchbox.client.resolvers import (
    Resolver,
    ResolverMethod,
    ResolverSettings,
    add_resolver_class,
)
from matchbox.common.arrow import SCHEMA_CLUSTERS
from matchbox.common.dtos import (
    ModelStepName,
    ResolverStepName,
    ResolverType,
    SourceStepName,
)
from matchbox.common.factories.entities import ClusterEntity, SourceEntity
from matchbox.common.factories.models import ModelTestkit, model_factory
from matchbox.common.factories.sources import linked_sources_factory
from matchbox.common.transform import DisjointSet


class MockResolverSettings(ResolverSettings):
    """Settings type for MockResolver."""

    thresholds: dict[
        ModelStepName,
        Annotated[float, Field(ge=0.0, le=1.0)],
    ] = Field(default_factory=dict)

    def validate_inputs(self, model_names: Iterable[ModelStepName]) -> None:
        """Validate all model names are present in thresholds."""
        if missing := set(model_names) - set(self.thresholds.keys()):
            raise RuntimeError(f"Missing thresholds for models: {missing}")


def _connected_components_from_edges(
    model_edges: Mapping[ModelStepName, pl.DataFrame],
    thresholds: Mapping[ModelStepName, float],
) -> pl.DataFrame:
    """Generate clusters from model edge tables and per-model thresholds."""
    djs = DisjointSet[int]()

    for model_name, edges in model_edges.items():
        if edges.height == 0:
            continue

        threshold = thresholds[model_name]
        filtered_edges = edges.filter(pl.col("score") >= threshold)
        for left_id, right_id in filtered_edges.select(
            "left_id", "right_id"
        ).iter_rows():
            djs.union(left_id, right_id)

    rows: list[dict[str, int]] = []
    for parent_id, component in enumerate(djs.get_components(), start=1):
        rows.extend(
            {"parent_id": parent_id, "child_id": node_id} for node_id in component
        )

    if not rows:
        return pl.from_arrow(SCHEMA_CLUSTERS.empty_table())
    return pl.DataFrame(rows).cast(pl.Schema(SCHEMA_CLUSTERS))


class MockResolver(ResolverMethod):
    """Mock resolver methodology used by resolver testkits."""

    resolver_type: ClassVar[ResolverType] = ResolverType.COMPONENTS
    settings: MockResolverSettings

    def compute_clusters(
        self, model_edges: Mapping[ModelStepName, pl.DataFrame]
    ) -> pl.DataFrame:
        """Compute mock clusters with connected components."""
        self.settings.validate_inputs(model_edges.keys())
        return _connected_components_from_edges(
            model_edges=model_edges,
            thresholds=self.settings.thresholds,
        )


add_resolver_class(MockResolver)


class ResolverTestkit(BaseModel):
    """Resolver plus local expected data for tests."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    resolver: Resolver
    assignments: pl.DataFrame
    entities: tuple[ClusterEntity, ...]

    @property
    def name(self) -> str:
        """Return resolver name."""
        return self.resolver.name

    def query(self) -> Query:
        """Thin wrapper to Query this testkit's Sources via its Resolver."""
        return self.resolver.query()

    def fake_run(self) -> Self:
        """Set resolver results without running the resolver."""
        self.resolver.results = self.assignments
        return self

    def into_dag(self) -> dict[str, Any]:
        """Return kwargs for explicit DAG insertion."""
        config = self.resolver.config
        return {
            "name": self.resolver.name,
            "inputs": list(config.inputs),
            "resolver_class": config.resolver_class,
            "resolver_settings": config.resolver_settings,
            "description": self.resolver.description,
        }


def resolver_factory(
    dag: DAG | None = None,
    inputs: Iterable[ModelTestkit] | None = None,
    true_entities: Iterable[SourceEntity] | None = None,
    name: ResolverStepName | None = None,
    description: str | None = None,
    thresholds: Mapping[ModelStepName, float] | None = None,
    seed: int = 42,
) -> ResolverTestkit:
    """Generate a complete resolver testkit.

    Allows autoconfiguration with minimal settings, or more nuanced control.

    Can either be used to generate a resolver in a pipeline, interconnected with
    existing testkit objects, or generate a standalone resolver with random data.

    Args:
        dag: DAG containing this resolver.
            Inferred from the first input testkit if not provided.
            A default DAG is created when inputs are also absent.
        inputs: An iterable of ModelTestkit objects to use as resolver inputs.
            If None, a single default deduper model testkit is created automatically.
            All inputs must belong to the same DAG.
        true_entities: Ground truth SourceEntity objects used to generate the
            expected cluster assignments. If None, the resolver testkit will have
            no expected entities.
        name: Name of the resolver. Defaults to a randomly generated word suffixed
            with '_resolver'.
        description: Description of the resolver.
        thresholds: Per-model score thresholds in [0.0, 1.0]. If omitted,
            defaults to 0.0 for all resolver inputs.
        seed: Random seed for reproducibility.

    Returns:
        ResolverTestkit: A resolver testkit with generated assignments and expected
            entities.

    Raises:
        TypeError: If any element of inputs is not a ModelTestkit.
        ValueError: If inputs belong to different DAGs.
    """
    if inputs is None:
        if dag is None:
            dag = DAG(name="collection")
            dag.run = 1
        linked = linked_sources_factory(dag=dag, seed=seed)
        default_model = model_factory(
            left_testkit=linked.sources["crn"],
            true_entities=tuple(linked.true_entities),
            seed=seed,
        ).fake_run()
        inputs = (default_model,)
        source_names: set[SourceStepName] = set(linked.sources.keys())
    else:
        inputs = tuple(inputs)
        source_names: set[SourceStepName] = set()
        for testkit in inputs:
            if not isinstance(testkit, ModelTestkit):
                raise TypeError("resolver_factory inputs must be ModelTestkit.")
            source_names.update(testkit.model.sources)
        dag = dag or inputs[0].model.dag

    input_map: dict[str, ModelTestkit] = {}
    for testkit in inputs:
        input_map.setdefault(testkit.name, testkit)

    resolver_inputs: list[Model] = []
    model_edges: dict[ModelStepName, pl.DataFrame] = {}
    for testkit in input_map.values():
        if testkit.model.dag != dag:
            raise ValueError("Cannot mix DAGs when building a resolver testkit.")
        if testkit.model.results is None:
            testkit.fake_run()
        resolver_inputs.append(testkit.model)
        model_edges[testkit.name] = testkit.scores

    expected_model_names = set(input_map.keys())
    if thresholds is None:
        thresholds = {name: 0.0 for name in expected_model_names}

    if set(thresholds) != expected_model_names:
        raise ValueError("Threshold keys must exactly match resolver input models.")

    resolver_settings = MockResolverSettings(
        thresholds=thresholds,
    )

    generator = Faker()
    generator.seed_instance(seed)
    resolver = Resolver(
        dag=dag,
        name=name or f"{generator.unique.word()}_resolver",
        inputs=resolver_inputs,
        resolver_class=MockResolver,
        resolver_settings=resolver_settings,
        description=description,
    )
    assignments = _connected_components_from_edges(
        model_edges=model_edges,
        thresholds=thresholds,
    )

    source_entities = tuple(true_entities or ())
    entities = tuple(
        projected
        for entity in source_entities
        if (projected := entity.to_cluster_entity(*source_names)) is not None
    )

    return ResolverTestkit(
        resolver=resolver,
        assignments=assignments,
        entities=entities,
    )



from io import BytesIO

from pyarrow import Table
from pyarrow.parquet import read_table

from matchbox.client._handler.main import CLIENT, http_retry, url_params
from matchbox.common.arrow import (
    SCHEMA_QUERY,
    SCHEMA_QUERY_WITH_LEAVES,
    check_schema_subset,
)
from matchbox.common.dtos import (
    Match,
    ResolverStepPath,
    SourceStepPath,
)
from matchbox.common.exceptions import MatchboxEmptyServerResponse
from matchbox.common.logging import logger


@http_retry
def query(
    source: SourceStepPath,
    return_leaf_id: bool,
    resolver: ResolverStepPath | None = None,
    limit: int | None = None,
) -> Table:
    """Query a source in Matchbox."""
    log_prefix = f"Query {source}"
    logger.debug(f"Using {resolver}", prefix=log_prefix)

    params = url_params(
        {
            "collection": source.collection,
            "run_id": source.run,
            "source": source.name,
            "resolver": resolver.name if resolver else None,
            "return_leaf_id": return_leaf_id,
            "limit": limit,
        }
    )
    res = CLIENT.get("/query", params=params)

    buffer = BytesIO(res.content)
    table = read_table(buffer)

    logger.debug("Finished", prefix=log_prefix)

    expected_schema = SCHEMA_QUERY_WITH_LEAVES if return_leaf_id else SCHEMA_QUERY

    check_schema_subset(expected_schema, table.schema)

    if table.num_rows == 0:
        raise MatchboxEmptyServerResponse(operation="query")

    return table


@http_retry
def match(
    targets: list[SourceStepPath],
    source: SourceStepPath,
    key: str,
    resolver: ResolverStepPath,
) -> list[Match]:
    """Match a source against a list of targets."""
    log_prefix = f"Query {source}"
    target_names = ", ".join(str(target) for target in targets)
    logger.debug(
        f"{key} to {target_names} using {resolver}",
        prefix=log_prefix,
    )

    params = url_params(
        {
            "collection": resolver.collection,
            "run_id": resolver.run,
            "targets": [t.name for t in targets],
            "source": source.name,
            "key": key,
            "resolver": resolver.name,
        }
    )
    res = CLIENT.get("/match", params=params)

    logger.debug("Finished", prefix=log_prefix)

    matches = [Match.model_validate(m) for m in res.json()]

    if not matches:
        raise MatchboxEmptyServerResponse(operation="match")

    return matches



"""Functions to transform data between tabular and graph structures."""

from collections import defaultdict
from collections.abc import Hashable
from typing import Generic, TypeVar

from matchbox.common.hash import HASH_FUNC

T = TypeVar("T", bound=Hashable)


class DisjointSet(Generic[T]):
    """Disjoint set forest with "path compression" and "union by rank" heuristics.

    This follows implementation from Cormen, Thomas H., et al. Introduction to
    algorithms. MIT press, 2022
    """

    def __init__(self) -> None:
        """Initialize the disjoint set."""
        self.parent: dict[T, T] = {}
        self.rank: dict[T, int] = {}

    def _make_set(self, x: T) -> None:
        """Create a new set with a single element x."""
        self.parent[x] = x
        self.rank[x] = 0

    def add(self, x: T) -> None:
        """Add a new element to the disjoint set."""
        if x not in self.parent:
            self._make_set(x)

    def union(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        self._link(self._find(x), self._find(y))

    def _link(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        if self.rank[x] > self.rank[y]:
            self.parent[y] = x
        else:
            self.parent[x] = y
            if self.rank[x] == self.rank[y]:
                self.rank[y] += 1

    def _find(self, x: T) -> T:
        """Return the representative element of the set containing x."""
        if x not in self.parent:
            self._make_set(x)
            return x

        if x != self.parent[x]:
            self.parent[x] = self._find(self.parent[x])

        return self.parent[x]

    def get_components(self) -> list[set[T]]:
        """Return the connected components of the disjoint set."""
        components = defaultdict(set)
        for x in self.parent:
            root = self._find(x)
            components[root].add(x)
        return list(components.values())


def hash_cluster_leaves(leaves: list[bytes]) -> bytes:
    """Canonical method to convert list of cluster IDs to their combined hash."""
    return HASH_FUNC(b"|".join(leaf for leaf in sorted(leaves))).digest()



"""Functions and classes to define, run and register models."""

import inspect
from typing import TYPE_CHECKING, Any, ClassVar, ParamSpec, TypeVar, overload

import polars as pl
import pyarrow as pa
from polars import DataFrame

from matchbox.client.models import dedupers, linkers
from matchbox.client.models.dedupers.base import Deduper, DeduperSettings
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.client.queries import Query
from matchbox.client.results import normalise_model_scores
from matchbox.client.steps import StepABC, post_run
from matchbox.common.arrow import SCHEMA_MODEL_EDGES
from matchbox.common.dtos import (
    ModelConfig,
    ModelStepName,
    ModelStepPath,
    ModelType,
    SourceStepName,
    Step,
    StepType,
)
from matchbox.common.hash import hash_arrow_table
from matchbox.common.logging import logger, profile_time

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.resolvers import Resolver
    from matchbox.client.resolvers.base import ResolverMethod, ResolverSettings
else:
    DAG = Any

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")


_MODEL_CLASSES = {
    **{name: obj for name, obj in inspect.getmembers(dedupers, inspect.isclass)},
    **{name: obj for name, obj in inspect.getmembers(linkers, inspect.isclass)},
}


def add_model_class(ModelClass: type[Linker] | type[Deduper]) -> None:
    """Add custom deduper or linker."""
    if issubclass(ModelClass, Linker) or issubclass(ModelClass, Deduper):
        _MODEL_CLASSES[ModelClass.__name__] = ModelClass
    else:
        raise ValueError("The argument is not a proper subclass of Deduper or Linker.")


class Model(StepABC):
    """Unified model class for both linking and deduping operations."""

    _local_data_schema: ClassVar[pa.Schema] = SCHEMA_MODEL_EDGES

    @overload
    def __init__(
        self,
        name: str,
        dag: DAG,
        model_class: type[Deduper],
        model_settings: DeduperSettings | dict,
        left_query: Query,
        right_query: None = None,
        description: str | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Linker],
        model_settings: LinkerSettings | dict,
        left_query: Query,
        right_query: Query,
        description: str | None = None,
    ) -> None: ...

    def __init__(
        self,
        dag: DAG,
        name: str,
        model_class: type[Deduper] | type[Linker] | str,
        model_settings: DeduperSettings | LinkerSettings | dict,
        left_query: Query,
        right_query: Query | None = None,
        description: str | None = None,
    ):
        """Create a new model instance.

        Args:
            dag: DAG containing this model.
            name: Unique name for the model.
            model_class: Class of Linker or Deduper, or its name.
            model_settings: Appropriate settings object to pass to model class.
            left_query: The query that will get the data to deduplicate, or the data to
                link on the left.
            right_query: The query that will get the data to link on the right.
            description: Optional description of the model.
        """
        super().__init__(dag=dag, name=name, description=description)

        self.left_query = left_query
        self.right_query = right_query

        if isinstance(model_class, str):
            self.model_class: type[Linker | Deduper] = _MODEL_CLASSES[model_class]
        else:
            self.model_class = model_class
        self.model_instance = self.model_class(settings=model_settings)

        self.model_type: ModelType = (
            ModelType.LINKER
            if issubclass(self.model_class, Linker)
            else ModelType.DEDUPER
        )

        if isinstance(model_settings, dict):
            SettingsClass = self.model_instance.__annotations__["settings"]
            self.model_settings = SettingsClass(**model_settings)
        else:
            self.model_settings = model_settings

    @property
    def results(self) -> pl.DataFrame | None:
        """The locally computed model scores. Alias for local_data."""
        return self._local_data

    @results.setter
    def results(self, value: pl.DataFrame | None) -> None:
        self._local_data = value

    @property
    def config(self) -> ModelConfig:
        """Generate config DTO from Model."""
        return ModelConfig(
            type=self.model_type,
            model_class=self.model_class.__name__,
            model_settings=self.model_settings.model_dump(mode="json"),
            left_query=self.left_query.config,
            right_query=self.right_query.config if self.right_query else None,
        )

    @property
    def sources(self) -> set[SourceStepName]:
        """Set of source names upstream of this node."""
        left_input = self.dag.nodes[self.left_query.config.resolves_from]
        model_sources = left_input.sources
        if self.right_query:
            right_input = self.dag.nodes[self.right_query.config.resolves_from]
            model_sources.update(right_input.sources)
        return model_sources

    @post_run
    def _fingerprint(self) -> bytes:
        """Compute a content hash invariant to left/right ID order."""
        return hash_arrow_table(
            self._local_data.to_arrow(), as_sorted_list=["left_id", "right_id"]
        )

    @post_run
    def to_dto(self) -> Step:
        """Convert to Step DTO for API calls."""
        return Step(
            description=self.description,
            step_type=StepType.MODEL,
            config=self.config,
            fingerprint=self._fingerprint(),
        )

    @classmethod
    def from_dto(
        cls,
        step: Step,
        step_name: str,
        dag: DAG,
        **kwargs: Any,
    ) -> "Model":
        """Reconstruct from Step DTO."""
        if step.step_type != StepType.MODEL:
            raise ValueError("Step must be of type 'model'")

        return cls(
            dag=dag,
            name=ModelStepName(step_name),
            description=step.description,
            model_class=step.config.model_class,
            model_settings=step.config.model_settings,
            left_query=Query.from_config(step.config.left_query, dag=dag),
            right_query=Query.from_config(step.config.right_query, dag=dag)
            if step.config.right_query
            else None,
        )

    @property
    def path(self) -> ModelStepPath:
        """Return the model step path."""
        return ModelStepPath(
            collection=self.dag.name,
            run=self.dag.run,
            name=self.name,
        )

    @profile_time(attr="name")
    def compute_scores(
        self, left_df: DataFrame, right_df: DataFrame | None = None
    ) -> DataFrame:
        """Run model instance against data."""
        if self.config.type == ModelType.LINKER:
            self.model_instance.prepare(left_df, right_df)
            scores = self.model_instance.link(left=left_df, right=right_df)
        else:
            self.model_instance.prepare(left_df)
            scores = self.model_instance.dedupe(data=left_df)
        return scores

    def run(
        self,
        left_data: DataFrame | None = None,
        right_data: DataFrame | None = None,
        low_memory: bool = False,
    ) -> pl.DataFrame:
        """Execute the model pipeline and return results.

        Args:
            left_data (optional): Pre-fetched query data to deduplicate if the model is
                a deduper, or link on the left if the model is a linker.
            right_data (optional): Pre-fetched query data to link on the right, if the
                model is a linker. If the model is a deduper, this argument is ignored.
            low_memory: If True, it will not download data from the server to support
                evaluation.
        """
        log_prefix = f"Run {self.name}"
        logger.info("Executing left query", prefix=log_prefix)

        left_df = (
            left_data
            if left_data is not None
            else self.left_query.data(cache_leaf_ids=(not low_memory))
        )
        right_df = None

        if self.config.type == ModelType.LINKER:
            logger.info("Executing right query", prefix=log_prefix)
            right_df = (
                right_data
                if right_data is not None
                else self.right_query.data(cache_leaf_ids=(not low_memory))
            )

        logger.info("Running model logic", prefix=log_prefix)
        scores = self.compute_scores(left_df, right_df)
        self._local_data = normalise_model_scores(scores)

        return self._local_data

    def resolver(
        self,
        *other_models: "Model",
        name: str,
        resolver_class: type["ResolverMethod"] | str,
        resolver_settings: "ResolverSettings | dict[str, Any] | None" = None,
        description: str | None = None,
    ) -> "Resolver":
        """Create a resolver rooted at this model and add it to the DAG."""
        return self.dag.resolver(
            name=name,
            inputs=[self, *other_models],
            resolver_class=resolver_class,
            resolver_settings=resolver_settings if resolver_settings else {},
            description=description,
        )

    def clear_data(self) -> None:
        """Suppress data clearing for models."""
        # TODO make DAG smart enough to know when to call this
        # without affecting downstream resolvers



"""Tests for the Components resolver methodology."""

import polars as pl
import pytest

from matchbox.client.resolvers import Components, ComponentsSettings
from matchbox.common.arrow import SCHEMA_CLUSTERS


def test_components_compute_clusters_uses_thresholds() -> None:
    """Test thresholds are honoured by the Components.compute_clusters."""
    method = Components(settings=ComponentsSettings(thresholds={"model_a": 0.6}))
    model_edges = {
        "model_a": pl.DataFrame(
            {
                "left_id": [1, 2],
                "right_id": [2, 3],
                "score": [0.8, 0.4],
            },
            schema={
                "left_id": pl.UInt64,
                "right_id": pl.UInt64,
                "score": pl.Float32,
            },
        )
    }

    clusters = method.compute_clusters(model_edges=model_edges)

    grouped_clusters = {
        frozenset(group["child_id"].to_list())
        for group in clusters.partition_by("parent_id")
    }
    assert grouped_clusters == {frozenset({1, 2})}


def test_components_compute_clusters_returns_empty_for_no_edges() -> None:
    """Test Components.compute_clusters can work with no data."""
    clusters = Components(settings=ComponentsSettings()).compute_clusters(
        model_edges={}
    )
    assert clusters.height == 0
    assert clusters.schema == pl.Schema(SCHEMA_CLUSTERS)


def test_components_compute_clusters_merges_multiple_models() -> None:
    """Test Components.compute_clusters can work with multiple models."""
    method = Components(
        settings=ComponentsSettings(
            thresholds={"model_a": 0.0}  # one threshold implicit
        )
    )
    model_edges = {
        "model_a": pl.DataFrame(
            {"left_id": [1], "right_id": [2], "score": [0.9]},
            schema={
                "left_id": pl.UInt64,
                "right_id": pl.UInt64,
                "score": pl.Float32,
            },
        ),
        "model_b": pl.DataFrame(
            {"left_id": [3], "right_id": [4], "score": [0.8]},
            schema={
                "left_id": pl.UInt64,
                "right_id": pl.UInt64,
                "score": pl.Float32,
            },
        ),
    }

    clusters = method.compute_clusters(model_edges=model_edges)

    grouped_clusters = {
        frozenset(group["child_id"].to_list())
        for group in clusters.partition_by("parent_id")
    }
    assert grouped_clusters == {frozenset({1, 2}), frozenset({3, 4})}


def test_components_compute_clusters_errors_on_unknown_model() -> None:
    """Test that a threshold for an unknown model raises a RuntimeError."""
    method = Components(settings=ComponentsSettings(thresholds={"fooo": 0.5}))

    with pytest.raises(RuntimeError, match="Unknown models in thresholds"):
        method.compute_clusters(model_edges={"foo": pl.DataFrame()})



"""ORM classes for the Matchbox PostgreSQL database."""

import textwrap
from typing import Any, Optional

from sqlalchemy import (
    BIGINT,
    BOOLEAN,
    INTEGER,
    REAL,
    CheckConstraint,
    DateTime,
    Enum,
    ForeignKey,
    Identity,
    Index,
    UniqueConstraint,
    select,
    text,
)
from sqlalchemy.dialects.postgresql import BYTEA, JSONB, TEXT, insert
from sqlalchemy.orm import Mapped, Session, mapped_column, relationship, selectinload

from matchbox.common.dtos import (
    BackendResourceType,
    CollectionName,
    LocationConfig,
    ModelType,
    RunID,
    StepName,
    StepPath,
    StepType,
    UploadStage,
)
from matchbox.common.dtos import Collection as CommonCollection
from matchbox.common.dtos import ModelConfig as CommonModelConfig
from matchbox.common.dtos import ResolverConfig as CommonResolverConfig
from matchbox.common.dtos import Run as CommonRun
from matchbox.common.dtos import SourceConfig as CommonSourceConfig
from matchbox.common.dtos import SourceField as CommonSourceField
from matchbox.common.dtos import Step as CommonStep
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxRunNotFoundError,
    MatchboxStepAlreadyExists,
    MatchboxStepNotFoundError,
    MatchboxStepTypeError,
)
from matchbox.server.base import (
    DEFAULT_GROUPS,
    DEFAULT_PERMISSIONS,
)
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.mixin import CountMixin


class Collections(CountMixin, MBDB.MatchboxBase):
    """Named collections of steps and runs."""

    __tablename__ = "collections"

    collection_id: Mapped[int] = mapped_column(
        BIGINT, primary_key=True, autoincrement=True
    )
    name: Mapped[str] = mapped_column(TEXT, nullable=False)

    # Relationships
    runs: Mapped[list["Runs"]] = relationship(back_populates="collection")
    permissions: Mapped[list["Permissions"]] = relationship(
        back_populates="collection",
        passive_deletes=True,
    )

    # Constraints
    __table_args__ = (UniqueConstraint("name", name="collections_name_key"),)

    @classmethod
    def from_name(
        cls,
        name: CollectionName,
        session: Session | None = None,
    ) -> "Collections":
        """Resolve a collection name to a Collections object.

        Args:
            name: The name of the collection to resolve.
            session: Optional session to use for the query.

        Raises:
            MatchboxCollectionNotFoundError: If the collection doesn't exist.
        """
        query = select(cls).where(cls.name == name).options(selectinload(cls.runs))

        if session:
            collection = session.execute(query).scalar_one_or_none()
        else:
            with MBDB.get_session() as session:
                collection = session.execute(query).scalar_one_or_none()

        if not collection:
            raise MatchboxCollectionNotFoundError(f"Collection '{name}' not found.")

        return collection

    def to_dto(self) -> CommonCollection:
        """Convert ORM collection to a matchbox.common Collection object."""
        run_ids: list[RunID] = []
        default_run = None
        if runs := self.runs:
            run_ids = [r.run_id for r in runs]
            default_run_list = [r.run_id for r in runs if r.is_default]
            if default_run_list:
                default_run = default_run_list[0]

        return CommonCollection(runs=run_ids, default_run=default_run)


class Runs(CountMixin, MBDB.MatchboxBase):
    """Runs of collections of steps."""

    __tablename__ = "runs"

    run_id: Mapped[int] = mapped_column(BIGINT, primary_key=True, autoincrement=True)
    collection_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("collections.collection_id", ondelete="CASCADE"),
        nullable=False,
    )
    is_mutable: Mapped[bool] = mapped_column(BOOLEAN, default=False, nullable=True)
    is_default: Mapped[bool] = mapped_column(BOOLEAN, default=False, nullable=True)

    # Relationships
    collection: Mapped["Collections"] = relationship(back_populates="runs")
    steps: Mapped[list["Steps"]] = relationship(back_populates="run")

    # Constraints
    __table_args__ = (
        UniqueConstraint("collection_id", "run_id", name="unique_run_id"),
        Index(
            "ix_default_run_collection",
            "collection_id",
            unique=True,
            postgresql_where=text("is_default = true"),
        ),
    )

    @classmethod
    def from_id(
        cls,
        collection: CollectionName,
        run_id: RunID,
        session: Session | None = None,
    ) -> "Runs":
        """Resolve a collection and run name to a Runs object.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run within that collection.
            session: Optional session to use for the query.

        Raises:
            MatchboxRunNotFoundError: If the run doesn't exist.
        """
        query = (
            select(cls)
            .where(cls.run_id == run_id)
            .options(
                selectinload(cls.steps)
                .selectinload(Steps.source_config)
                .selectinload(SourceConfigs.fields),
                selectinload(cls.steps).selectinload(Steps.model_config),
                selectinload(cls.steps).selectinload(Steps.resolver_config),
                selectinload(cls.collection),
            )
        )

        if session:
            run_orm = session.execute(query).scalar_one_or_none()
        else:
            with MBDB.get_session() as session:
                run_orm = session.execute(query).scalar_one_or_none()

        if not run_orm:
            raise MatchboxRunNotFoundError

        if run_orm.collection.name != collection:
            raise MatchboxRunNotFoundError(
                run_id=id,
                message=f"Run {id} not found in collection {collection}",
            )

        return run_orm

    def to_dto(self) -> CommonRun:
        """Convert ORM run to a matchbox.common Run object."""
        steps: dict[StepName, CommonStep] = {}
        if self.steps:
            steps = {step.name: step.to_dto() for step in self.steps}

        return CommonRun(
            run_id=self.run_id,
            is_default=self.is_default,
            is_mutable=self.is_mutable,
            steps=steps,
        )


class StepFrom(CountMixin, MBDB.MatchboxBase):
    """Step lineage closure table."""

    __tablename__ = "step_from"

    # Columns
    parent: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        primary_key=True,
    )
    child: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        primary_key=True,
    )
    level: Mapped[int] = mapped_column(INTEGER, nullable=False, primary_key=True)

    # Constraints
    __table_args__ = (
        CheckConstraint("parent != child", name="no_self_reference"),
        CheckConstraint("level > 0", name="positive_level"),
    )


class Steps(CountMixin, MBDB.MatchboxBase):
    """Table of steps corresponding to models, resolvers, and sources.

    Models produce edges and resolvers produce cluster assignments.
    """

    __tablename__ = "steps"

    # Columns
    step_id: Mapped[int] = mapped_column(BIGINT, primary_key=True, autoincrement=True)
    run_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("runs.run_id", ondelete="CASCADE"), nullable=False
    )
    upload_stage: Mapped[UploadStage] = mapped_column(
        Enum(UploadStage, native_enum=True, name="upload_stages", schema="mb"),
        nullable=False,
        default=UploadStage.READY,
    )
    name: Mapped[str] = mapped_column(TEXT, nullable=False)
    description: Mapped[str | None] = mapped_column(TEXT, nullable=True)
    type: Mapped[str] = mapped_column(TEXT, nullable=False)
    fingerprint: Mapped[bytes] = mapped_column(BYTEA, nullable=False)

    # Relationships
    source_config: Mapped[Optional["SourceConfigs"]] = relationship(
        back_populates="source_step", uselist=False
    )
    model_config: Mapped[Optional["ModelConfigs"]] = relationship(
        back_populates="model_step", uselist=False
    )
    resolver_config: Mapped[Optional["ResolverConfigs"]] = relationship(
        back_populates="resolver_step", uselist=False
    )
    model_edges: Mapped[list["ModelEdges"]] = relationship(
        back_populates="proposed_by",
        passive_deletes=True,
    )
    resolver_clusters: Mapped[list["ResolverClusters"]] = relationship(
        back_populates="proposed_by",
        passive_deletes=True,
    )
    parents: Mapped[list["Steps"]] = relationship(
        secondary=StepFrom.__table__,
        primaryjoin=textwrap.dedent("""\
            and_(
                Steps.step_id == StepFrom.child,
                StepFrom.level == 1
            )
        """),
        secondaryjoin="Steps.step_id == StepFrom.parent",
        viewonly=True,
        order_by="StepFrom.parent",
    )
    run: Mapped["Runs"] = relationship(back_populates="steps")

    # Constraints
    __table_args__ = (
        CheckConstraint(
            "type IN ('model', 'source', 'resolver')",
            name="step_type_constraints",
        ),
        UniqueConstraint("run_id", "name", name="steps_name_key"),
    )

    @property
    def ancestors(self) -> set["Steps"]:
        """Return all ancestors (parents, grandparents, etc.) of this step."""
        with MBDB.get_session() as session:
            ancestor_query = (
                select(Steps)
                .select_from(Steps)
                .join(StepFrom, Steps.step_id == StepFrom.parent)
                .where(StepFrom.child == self.step_id)
            )
            return set(session.execute(ancestor_query).scalars().all())

    @property
    def descendants(self) -> set["Steps"]:
        """Return descendants (children, grandchildren, etc.) of this step."""
        with MBDB.get_session() as session:
            descendant_query = (
                select(Steps)
                .select_from(Steps)
                .join(StepFrom, Steps.step_id == StepFrom.child)
                .where(StepFrom.parent == self.step_id)
            )
            return set(session.execute(descendant_query).scalars().all())

    def get_lineage(
        self,
        sources: list["SourceConfigs"] | None = None,
        queryable_only: bool = False,
    ) -> list[tuple[int, int | None]]:
        """Returns lineage ordered by priority.

        Highest priority (lowest level) first, then by step_id for stability.

        Args:
            sources: If provided, only return lineage paths that lead to these sources
            queryable_only: If true, only include queryable step types

        Returns:
            List of tuples (step_id, source_config_id) ordered by priority.
        """
        with MBDB.get_session() as session:
            query = (
                select(
                    StepFrom.parent,
                    SourceConfigs.source_config_id,
                )
                .join(
                    Steps,
                    StepFrom.parent == Steps.step_id,
                )
                .join(
                    SourceConfigs,
                    StepFrom.parent == SourceConfigs.step_id,
                    isouter=True,
                )
                .where(StepFrom.child == self.step_id)
            )

            if queryable_only:
                query = query.where(
                    Steps.type.in_(
                        [
                            StepType.SOURCE.value,
                            StepType.RESOLVER.value,
                        ]
                    )
                )

            if sources:
                # Filter by source configs
                source_step_ids = [sc.step_id for sc in sources]

                descendant_ids = (
                    session.execute(
                        select(StepFrom.child).where(
                            StepFrom.parent.in_(source_step_ids)
                        )
                    )
                    .scalars()
                    .all()
                )

                query = query.where(
                    StepFrom.parent.in_(source_step_ids + descendant_ids)
                )

            results = session.execute(
                query.order_by(StepFrom.level.asc(), StepFrom.parent.asc())
            ).all()

            # Get self's source config ID
            self_source_config_id = (
                self.source_config.source_config_id if self.source_config else None
            )

            # Add self at beginning (highest priority - level 0)
            lineage_self = []
            if not queryable_only or self.type in {
                StepType.SOURCE.value,
                StepType.RESOLVER.value,
            }:
                lineage_self = [(self.step_id, self_source_config_id)]

            return lineage_self + list(results)

    @classmethod
    def from_path(
        cls,
        path: StepPath,
        res_type: StepType | None = None,
        session: Session | None = None,
        for_update: bool = False,
    ) -> "Steps":
        """Resolve a step path to a Step ORM object.

        Args:
            path: The path of the step to resolve.
            res_type: A step type to use as filter.
            session: A session to get the step for updates.
            for_update: Locks the row until updated.

        Raises:
            MatchboxStepNotFoundError: If the step doesn't exist.
        """
        query = (
            select(cls)
            .join(cls.run)
            .join(Runs.collection)
            .where(
                cls.name == path.name,
                Runs.run_id == path.run,
                Collections.name == path.collection,
            )
        )

        if res_type:
            query = query.where(cls.type == res_type.value)

        if for_update:
            query = query.with_for_update(nowait=True)

        if session:
            step = session.execute(query).scalar()
        else:
            with MBDB.get_session() as session:
                step = session.execute(query).scalar()

        if step:
            return step

        raise MatchboxStepNotFoundError(
            message=f"No step {path} of type {res_type or 'any'}."
        )

    @classmethod
    def from_dto(cls, step: CommonStep, path: StepPath, session: Session) -> "Steps":
        """Create a Steps instance from a Step DTO object.

        The step will be added to the session and flushed (but not committed).

        For model steps, lineage entries will be created automatically.

        Args:
            step: The Step DTO to convert
            path: The full step path
            session: Database session (caller must commit)

        Returns:
            A Steps ORM instance with ID and relationships established
        """
        # Find the run ID for the given collection and run ID
        run_obj = session.execute(
            select(Runs)
            .join(Collections)
            .where(Collections.name == path.collection, Runs.run_id == path.run)
        ).scalar_one_or_none()

        if not run_obj:
            raise MatchboxRunNotFoundError(run_id=path.run)

        # Attempt to insert new step
        result = session.execute(
            insert(cls)
            .values(
                run_id=run_obj.run_id,
                name=path.name,
                description=step.description,
                type=step.step_type.value,
                fingerprint=step.fingerprint,
            )
            .on_conflict_do_nothing(constraint="steps_name_key")
            .returning(cls.step_id)
        )

        step_id = result.scalar_one_or_none()
        if step_id is None:
            raise MatchboxStepAlreadyExists(f"Step {path.name} already exists")

        # Fetch the newly created step
        step_orm = session.get(cls, step_id)

        if step.step_type == StepType.SOURCE:
            step_orm.source_config = SourceConfigs.from_dto(step.config)

        elif step.step_type == StepType.MODEL:
            step_orm.model_config = ModelConfigs.from_dto(step.config)
            # Create lineage
            left_parent_path = StepPath(
                collection=path.collection,
                run=path.run,
                name=step.config.left_query.resolves_from,
            )
            left_parent = cls.from_path(
                path=left_parent_path,
                session=session,
            )
            if left_parent.type not in {
                StepType.SOURCE.value,
                StepType.RESOLVER.value,
            }:
                raise MatchboxStepTypeError(
                    step_name=str(left_parent_path),
                    step_type=StepType(left_parent.type),
                    expected_step_types=[
                        StepType.SOURCE,
                        StepType.RESOLVER,
                    ],
                )
            cls._create_closure_entries(session, step_orm, left_parent)

            if step.config.type == ModelType.LINKER:
                right_parent_path = StepPath(
                    collection=path.collection,
                    run=path.run,
                    name=step.config.right_query.resolves_from,
                )
                right_parent = cls.from_path(
                    path=right_parent_path,
                    session=session,
                )
                if right_parent.type not in {
                    StepType.SOURCE.value,
                    StepType.RESOLVER.value,
                }:
                    raise MatchboxStepTypeError(
                        step_name=str(right_parent_path),
                        step_type=StepType(right_parent.type),
                        expected_step_types=[
                            StepType.SOURCE,
                            StepType.RESOLVER,
                        ],
                    )
                cls._create_closure_entries(session, step_orm, right_parent)

        elif step.step_type == StepType.RESOLVER:
            step_orm.resolver_config = ResolverConfigs.from_dto(step.config)
            for parent_name in dict.fromkeys(step.config.inputs):
                parent_path = StepPath(
                    collection=path.collection,
                    run=path.run,
                    name=parent_name,
                )
                parent = cls.from_path(
                    path=parent_path,
                    session=session,
                )
                if parent.type != StepType.MODEL.value:
                    raise MatchboxStepTypeError(
                        message=("Resolver steps must only depend on model steps."),
                        step_name=str(parent_path),
                        step_type=StepType(parent.type),
                        expected_step_types=[
                            StepType.MODEL,
                        ],
                    )
                cls._create_closure_entries(
                    session,
                    step_orm,
                    parent,
                )

        return step_orm

    def to_dto(self) -> CommonStep:
        """Convert ORM step to a matchbox.common Step object."""
        if self.type == StepType.SOURCE:
            config = self.source_config.to_dto()
        elif self.type == StepType.MODEL:
            config = self.model_config.to_dto()
        else:
            config = self.resolver_config.to_dto()

        return CommonStep(
            description=self.description,
            step_type=StepType(self.type),
            config=config,
            fingerprint=self.fingerprint,
        )

    @staticmethod
    def _create_closure_entries(
        session: Session, child: "Steps", parent: "Steps"
    ) -> None:
        """Create closure table entries for a parent-child relationship."""
        # Direct relationship
        session.execute(
            insert(StepFrom)
            .values(parent=parent.step_id, child=child.step_id, level=1)
            .on_conflict_do_nothing()
        )

        # Transitive closure
        ancestors = (
            session.execute(select(StepFrom).where(StepFrom.child == parent.step_id))
            .scalars()
            .all()
        )

        if ancestors:
            values_to_insert: list[dict[str, int]] = [
                {
                    "parent": ancestor.parent,
                    "child": child.step_id,
                    "level": ancestor.level + 1,
                }
                for ancestor in ancestors
            ]
            session.execute(
                insert(StepFrom).values(values_to_insert).on_conflict_do_nothing()
            )


class SourceFields(CountMixin, MBDB.MatchboxBase):
    """Table for storing column details for SourceConfigs."""

    __tablename__ = "source_fields"

    # Columns
    field_id: Mapped[int] = mapped_column(BIGINT, primary_key=True)
    source_config_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("source_configs.source_config_id", ondelete="CASCADE"),
        nullable=False,
    )
    index: Mapped[int] = mapped_column(INTEGER, nullable=False)
    name: Mapped[str] = mapped_column(TEXT, nullable=False)
    type: Mapped[str] = mapped_column(TEXT, nullable=False)
    is_key: Mapped[bool] = mapped_column(BOOLEAN, nullable=False)

    # Relationships
    source_config: Mapped["SourceConfigs"] = relationship(
        back_populates="fields",
        foreign_keys=[source_config_id],
    )

    # Constraints and indices
    __table_args__ = (
        UniqueConstraint("source_config_id", "index", name="unique_index"),
        Index("ix_source_columns_source_config_id", "source_config_id"),
        Index(
            "ix_unique_key_field",
            "source_config_id",
            unique=True,
            postgresql_where=text("is_key = true"),
        ),
    )


class ClusterSourceKey(CountMixin, MBDB.MatchboxBase):
    """Table for storing source primary keys for clusters."""

    __tablename__ = "cluster_keys"

    # Columns
    key_id: Mapped[int] = mapped_column(BIGINT, primary_key=True)
    cluster_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    source_config_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("source_configs.source_config_id", ondelete="CASCADE"),
        nullable=False,
    )
    key: Mapped[str] = mapped_column(TEXT, nullable=False)

    # Relationships
    cluster: Mapped["Clusters"] = relationship(back_populates="keys")
    source_config: Mapped["SourceConfigs"] = relationship(back_populates="cluster_keys")

    # Constraints and indices
    __table_args__ = (
        Index("ix_cluster_keys_cluster_id", "cluster_id"),
        Index("ix_cluster_keys_keys", "key"),
        Index("ix_cluster_keys_source_config_id", "source_config_id"),
        UniqueConstraint("key_id", "source_config_id", name="unique_keys_source"),
    )


class SourceConfigs(CountMixin, MBDB.MatchboxBase):
    """Table of source_configs of data for Matchbox."""

    __tablename__ = "source_configs"

    # Columns
    source_config_id: Mapped[int] = mapped_column(
        BIGINT, Identity(start=1), primary_key=True
    )
    step_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        nullable=False,
    )
    location_type: Mapped[str] = mapped_column(TEXT, nullable=False)
    location_name: Mapped[str] = mapped_column(TEXT, nullable=False)
    extract_transform: Mapped[str] = mapped_column(TEXT, nullable=False)

    @property
    def name(self) -> str:
        """Get the name of the related step."""
        return self.source_step.name

    # Relationships
    source_step: Mapped["Steps"] = relationship(back_populates="source_config")
    fields: Mapped[list["SourceFields"]] = relationship(
        back_populates="source_config",
        passive_deletes=True,
        cascade="all, delete-orphan",
    )
    key_field: Mapped[Optional["SourceFields"]] = relationship(
        primaryjoin=(
            "and_(SourceConfigs.source_config_id == SourceFields.source_config_id, "
            "SourceFields.is_key == True)"
        ),
        viewonly=True,
        uselist=False,
    )
    index_fields: Mapped[list["SourceFields"]] = relationship(
        primaryjoin=(
            "and_(SourceConfigs.source_config_id == SourceFields.source_config_id, "
            "SourceFields.is_key == False)"
        ),
        viewonly=True,
        order_by="SourceFields.index",
        collection_class=list,
    )
    cluster_keys: Mapped[list["ClusterSourceKey"]] = relationship(
        back_populates="source_config",
        passive_deletes=True,
    )
    clusters: Mapped[list["Clusters"]] = relationship(
        secondary=ClusterSourceKey.__table__,
        primaryjoin=(
            "SourceConfigs.source_config_id == ClusterSourceKey.source_config_id"
        ),
        secondaryjoin="ClusterSourceKey.cluster_id == Clusters.cluster_id",
        viewonly=True,
    )

    def __init__(
        self,
        key_field: SourceFields | None = None,
        index_fields: list[SourceFields] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise SourceConfigs with optional field objects."""
        super().__init__(**kwargs)

        # Add the key field and mark it as the key
        if key_field is not None:
            key_field.is_key = True
            key_field.index = 0
            self.fields.append(key_field)

        # Add index fields with proper indices
        if index_fields is not None:
            for idx, field in enumerate(index_fields):
                field.is_key = False
                field.index = idx + 1
                self.fields.append(field)

    @classmethod
    def list_all(cls) -> list["SourceConfigs"]:
        """Returns all source_configs in the database."""
        with MBDB.get_session() as session:
            return session.execute(select(cls)).scalars().all()

    @classmethod
    def from_dto(
        cls,
        config: CommonSourceConfig,
    ) -> "SourceConfigs":
        """Create a SourceConfigs instance from a Step DTO object."""
        # Create the SourceConfigs object
        return cls(
            location_type=str(config.location_config.type),
            location_name=str(config.location_config.name),
            extract_transform=config.extract_transform,
            key_field=SourceFields(
                index=0,
                name=config.key_field.name,
                type=config.key_field.type.value,
            ),
            index_fields=[
                SourceFields(
                    index=idx + 1,
                    name=field.name,
                    type=field.type.value,
                )
                for idx, field in enumerate(config.index_fields)
            ],
        )

    def to_dto(self) -> CommonSourceConfig:
        """Convert ORM source to a matchbox.common.SourceConfig object."""
        return CommonSourceConfig(
            location_config=LocationConfig(
                type=self.location_type, name=self.location_name
            ),
            extract_transform=self.extract_transform,
            key_field=CommonSourceField(
                name=self.key_field.name, type=self.key_field.type
            ),
            index_fields=[
                CommonSourceField(
                    name=field.name,
                    type=field.type,
                )
                for field in self.index_fields
            ],
        )


class ModelConfigs(CountMixin, MBDB.MatchboxBase):
    """Table of model configs for Matchbox."""

    __tablename__ = "model_configs"

    # Columns
    model_config_id: Mapped[int] = mapped_column(
        BIGINT, Identity(start=1), primary_key=True
    )
    step_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        nullable=False,
    )
    model_class: Mapped[str] = mapped_column(TEXT, nullable=False)
    model_settings: Mapped[dict[str, Any]] = mapped_column(JSONB, nullable=False)
    left_query: Mapped[dict[str, Any]] = mapped_column(JSONB, nullable=False)
    right_query: Mapped[dict[str, Any] | None] = mapped_column(JSONB, nullable=True)

    @property
    def name(self) -> str:
        """Get the name of the related step."""
        return self.model_step.name

    # Relationships
    model_step: Mapped["Steps"] = relationship(back_populates="model_config")

    @classmethod
    def list_all(cls) -> list["SourceConfigs"]:
        """Returns all model_configs in the database."""
        with MBDB.get_session() as session:
            return session.execute(select(cls)).scalars().all()

    @classmethod
    def from_dto(
        cls,
        config: CommonModelConfig,
    ) -> "ModelConfigs":
        """Create a SourceConfigs instance from a Step DTO object."""
        # Create the SourceConfigs object
        return cls(
            model_class=config.model_class,
            model_settings=config.model_settings,
            left_query=config.left_query.model_dump(mode="json"),
            right_query=(
                None
                if not config.right_query
                else config.right_query.model_dump(mode="json")
            ),
        )

    def to_dto(self) -> CommonModelConfig:
        """Convert ORM source to a matchbox.common.ModelConfig object."""
        return CommonModelConfig(
            type=ModelType.LINKER if self.right_query else ModelType.DEDUPER,
            model_class=self.model_class,
            model_settings=self.model_settings,
            left_query=self.left_query,
            right_query=self.right_query,
        )


class ResolverConfigs(CountMixin, MBDB.MatchboxBase):
    """Table of resolver configs for Matchbox."""

    __tablename__ = "resolver_configs"

    resolver_config_id: Mapped[int] = mapped_column(
        BIGINT, Identity(start=1), primary_key=True
    )
    step_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        nullable=False,
    )
    resolver_class: Mapped[str] = mapped_column(TEXT, nullable=False)
    resolver_settings: Mapped[dict[str, Any]] = mapped_column(JSONB, nullable=False)

    resolver_step: Mapped["Steps"] = relationship(back_populates="resolver_config")

    __table_args__ = (UniqueConstraint("step_id", name="resolver_configs_step_key"),)

    @classmethod
    def from_dto(
        cls,
        config: CommonResolverConfig,
    ) -> "ResolverConfigs":
        """Create a ResolverConfigs instance from a Step DTO object."""
        return cls(
            resolver_class=config.resolver_class,
            resolver_settings=config.resolver_settings,
        )

    def to_dto(self) -> CommonResolverConfig:
        """Convert ORM resolver config to a matchbox.common ResolverConfig object."""
        return CommonResolverConfig(
            resolver_class=self.resolver_class,
            resolver_settings=self.resolver_settings,
            inputs=tuple(r.name for r in self.resolver_step.parents),
        )


class Contains(CountMixin, MBDB.MatchboxBase):
    """Cluster lineage table."""

    __tablename__ = "contains"

    # Columns
    root: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )
    leaf: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )

    # Constraints and indices
    __table_args__ = (
        CheckConstraint("root != leaf", name="no_self_containment"),
        UniqueConstraint("root", "leaf"),
        Index("ix_contains_root_leaf", "root", "leaf"),
        Index("ix_contains_leaf_root", "leaf", "root"),
    )


class Clusters(CountMixin, MBDB.MatchboxBase):
    """Table of indexed data and clusters that match it."""

    __tablename__ = "clusters"

    # Columns
    cluster_id: Mapped[int] = mapped_column(BIGINT, primary_key=True)
    cluster_hash: Mapped[bytes] = mapped_column(BYTEA, nullable=False)

    # Relationships
    keys: Mapped[list["ClusterSourceKey"]] = relationship(
        back_populates="cluster",
        passive_deletes=True,
    )
    leaves: Mapped[list["Clusters"]] = relationship(
        secondary=Contains.__table__,
        primaryjoin="Clusters.cluster_id == Contains.root",
        secondaryjoin="Clusters.cluster_id == Contains.leaf",
        backref="roots",
    )
    # Add relationship to SourceConfigs through ClusterSourceKey
    source_configs: Mapped[list["SourceConfigs"]] = relationship(
        secondary=ClusterSourceKey.__table__,
        primaryjoin="Clusters.cluster_id == ClusterSourceKey.cluster_id",
        secondaryjoin=(
            "ClusterSourceKey.source_config_id == SourceConfigs.source_config_id"
        ),
        viewonly=True,
    )

    # Constraints and indices
    __table_args__ = (UniqueConstraint("cluster_hash", name="clusters_hash_key"),)


class UserGroups(MBDB.MatchboxBase):
    """Association table for user-group membership."""

    __tablename__ = "user_groups"

    # Columns
    user_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("users.user_id", ondelete="CASCADE"),
        primary_key=True,
    )
    group_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("groups.group_id", ondelete="CASCADE"),
        primary_key=True,
    )


class Users(CountMixin, MBDB.MatchboxBase):
    """Table of user identities."""

    __tablename__ = "users"

    # Columns
    user_id: Mapped[int] = mapped_column(BIGINT, primary_key=True)
    name: Mapped[str] = mapped_column(TEXT, nullable=False)
    email: Mapped[str] = mapped_column(TEXT, nullable=True)

    # Relationships
    judgements: Mapped[list["EvalJudgements"]] = relationship(back_populates="user")
    groups: Mapped[list["Groups"]] = relationship(
        secondary=UserGroups.__table__,
        back_populates="members",
    )

    # Constraints and indices
    __table_args__ = (UniqueConstraint("name", name="user_name_unique"),)


class Groups(CountMixin, MBDB.MatchboxBase):
    """Groups for permission management."""

    __tablename__ = "groups"

    # Columns
    group_id: Mapped[int] = mapped_column(BIGINT, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(TEXT, nullable=False)
    description: Mapped[str | None] = mapped_column(TEXT, nullable=True)
    is_system: Mapped[bool] = mapped_column(BOOLEAN, default=False, nullable=False)

    # Relationships
    members: Mapped[list["Users"]] = relationship(
        secondary=UserGroups.__table__,
        back_populates="groups",
    )
    permissions: Mapped[list["Permissions"]] = relationship(
        back_populates="group",
        passive_deletes=True,
    )

    # Constraints and indices
    __table_args__ = (UniqueConstraint("name", name="groups_name_key"),)

    @classmethod
    def initialise(cls) -> None:
        """Create standard users, groups, and permissions."""
        with MBDB.get_session() as session:
            # Upsert groups
            for group_dto in DEFAULT_GROUPS:
                session.execute(
                    insert(cls)
                    .values(
                        name=group_dto.name,
                        description=group_dto.description,
                        is_system=group_dto.is_system,
                    )
                    .on_conflict_do_nothing(index_elements=["name"])
                )

            # Collect and upsert all users
            all_users = {
                user_dto.user_name: user_dto
                for group_dto in DEFAULT_GROUPS
                for user_dto in (group_dto.members or [])
            }
            for user_dto in all_users.values():
                session.execute(
                    insert(Users)
                    .values(name=user_dto.user_name, email=user_dto.email)
                    .on_conflict_do_nothing(index_elements=["name"])
                )

            session.flush()

            # Cache lookups
            groups = {g.name: g for g in session.execute(select(cls)).scalars()}
            users = {u.name: u for u in session.execute(select(Users)).scalars()}

            # Upsert user-group memberships
            for group_dto in DEFAULT_GROUPS:
                group_obj = groups[group_dto.name]
                for user_dto in group_dto.members or []:
                    user_obj = users[user_dto.user_name]
                    session.execute(
                        insert(UserGroups)
                        .values(user_id=user_obj.user_id, group_id=group_obj.group_id)
                        .on_conflict_do_nothing()
                    )

            # Upsert permissions
            for grant, resource_type, resource_name in DEFAULT_PERMISSIONS:
                collection_id = None
                if resource_name and resource_type == BackendResourceType.COLLECTION:
                    collection = session.scalars(
                        select(Collections).where(Collections.name == resource_name)
                    ).one()
                    if not collection:
                        raise MatchboxCollectionNotFoundError(name=resource_name)
                    collection_id = collection.collection_id

                session.execute(
                    insert(Permissions)
                    .values(
                        group_id=groups[grant.group_name].group_id,
                        permission=grant.permission,
                        is_system=True
                        if resource_type == BackendResourceType.SYSTEM
                        else None,
                        collection_id=collection_id,
                    )
                    .on_conflict_do_nothing(constraint="unique_permission_grant")
                )

            session.commit()


class Permissions(CountMixin, MBDB.MatchboxBase):
    """Permissions granted to groups on resources.

    Each resource type should have one column. This creates lots of nulls,
    which are cheap in PostgreSQL and are on an ultimately small table,
    and avoids a polymorphic association.
    """

    __tablename__ = "permissions"

    # Columns
    permission_id: Mapped[int] = mapped_column(
        BIGINT, primary_key=True, autoincrement=True
    )
    permission: Mapped[str] = mapped_column(TEXT, nullable=False)
    group_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("groups.group_id", ondelete="CASCADE"),
        nullable=False,
    )
    collection_id: Mapped[int | None] = mapped_column(
        BIGINT,
        ForeignKey("collections.collection_id", ondelete="CASCADE"),
        nullable=True,
    )
    is_system: Mapped[bool | None] = mapped_column(
        BOOLEAN,
        nullable=True,
    )

    # Relationships
    group: Mapped["Groups"] = relationship(back_populates="permissions")
    collection: Mapped["Collections | None"] = relationship(
        back_populates="permissions"
    )

    # Constraints and indices
    __table_args__ = (
        CheckConstraint(
            "permission IN ('read', 'write', 'admin')",
            name="valid_permission",
        ),
        CheckConstraint(
            "(collection_id IS NOT NULL AND is_system IS NULL) OR "
            "(collection_id IS NULL AND is_system = true)",
            name="exactly_one_resource",
        ),
        UniqueConstraint(
            "permission",
            "group_id",
            "collection_id",
            "is_system",
            name="unique_permission_grant",
            postgresql_nulls_not_distinct=True,
        ),
    )


class EvalJudgements(CountMixin, MBDB.MatchboxBase):
    """Table of evaluation judgements produced by human validators."""

    __tablename__ = "eval_judgements"

    # Columns
    judgement_id: Mapped[int] = mapped_column(BIGINT, primary_key=True)
    user_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("users.user_id", ondelete="CASCADE"), nullable=False
    )
    endorsed_cluster_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    shown_cluster_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    tag: Mapped[str] = mapped_column(TEXT, nullable=True)
    timestamp: Mapped[DateTime] = mapped_column(DateTime(timezone=True), nullable=False)

    # Relationships
    user: Mapped["Users"] = relationship(back_populates="judgements")


class ModelEdges(CountMixin, MBDB.MatchboxBase):
    """Table of results for a model step.

    Stores the raw left/right scores created by a model.
    """

    __tablename__ = "model_edges"

    # Columns
    result_id: Mapped[int] = mapped_column(BIGINT, primary_key=True, autoincrement=True)
    step_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        nullable=False,
    )
    left_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    right_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    score: Mapped[float] = mapped_column(REAL, nullable=False)

    # Relationships
    proposed_by: Mapped["Steps"] = relationship(back_populates="model_edges")

    # Constraints
    __table_args__ = (
        Index("ix_model_edges_step", "step_id"),
        CheckConstraint(
            "score >= 0.0 AND score <= 1.0",
            name="valid_score",
        ),
        UniqueConstraint("step_id", "left_id", "right_id"),
    )


class ResolverClusters(CountMixin, MBDB.MatchboxBase):
    """Association table linking resolver steps to cluster IDs."""

    __tablename__ = "resolver_clusters"

    step_id: Mapped[int] = mapped_column(
        BIGINT,
        ForeignKey("steps.step_id", ondelete="CASCADE"),
        primary_key=True,
    )
    cluster_id: Mapped[int] = mapped_column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )

    proposed_by: Mapped["Steps"] = relationship(back_populates="resolver_clusters")

    __table_args__ = (Index("ix_resolver_clusters_step", "step_id"),)



"""Admin PostgreSQL mixin for Matchbox server."""

from typing import Literal, cast

from sqlalchemy import CursorResult, and_, bindparam, delete, select, union_all

from matchbox.common.dtos import (
    BackendResourceType,
    CollectionName,
    DefaultGroup,
    DefaultUser,
    GroupName,
    LoginResponse,
    PermissionGrant,
    PermissionType,
    User,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxDataNotFound,
    MatchboxDeletionNotConfirmed,
    MatchboxGroupNotFoundError,
)
from matchbox.common.logging import logger
from matchbox.server.base import (
    PERMISSION_GRANTS,
    MatchboxSnapshot,
)
from matchbox.server.postgresql.db import (
    MBDB,
    MatchboxBackends,
    MatchboxPostgresSettings,
)
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Collections,
    EvalJudgements,
    Groups,
    ModelEdges,
    Permissions,
    ResolverClusters,
    UserGroups,
    Users,
    insert,
)
from matchbox.server.postgresql.utils.db import dump, grant_permission, restore


class MatchboxPostgresAdminMixin:
    """Admin mixin for the PostgreSQL adapter for Matchbox."""

    settings: MatchboxPostgresSettings

    # User management

    def login(self, user: User) -> LoginResponse:  # noqa: D102
        with MBDB.get_session() as session:
            # Get public and admins groups
            public_group = session.scalars(
                select(Groups).where(Groups.name == DefaultGroup.PUBLIC)
            ).one()
            admins_group = session.scalars(
                select(Groups).where(Groups.name == DefaultGroup.ADMINS)
            ).one()

            # Upsert user
            if user.email:
                session.execute(
                    insert(Users)
                    .values(
                        name=user.user_name,
                        email=user.email,
                    )
                    .on_conflict_do_update(
                        index_elements=["name"],
                        set_={"email": user.email},
                    )
                )
            else:
                session.execute(
                    insert(Users)
                    .values(name=user.user_name)
                    .on_conflict_do_nothing(index_elements=["name"])
                )

            # Get the user object
            user_obj = session.scalars(
                select(Users).where(Users.name == user.user_name)
            ).one()

            # Ensure user is in public group
            session.execute(
                insert(UserGroups)
                .values(
                    user_id=user_obj.user_id,
                    group_id=public_group.group_id,
                )
                .on_conflict_do_nothing()
            )

            # Check if any other non-public user exists
            other_user_exists = (
                session.scalar(
                    select(Users.user_id)
                    .where(Users.name != DefaultUser.PUBLIC)
                    .where(Users.user_id != user_obj.user_id)
                    .limit(1)
                )
                is not None
            )

            setup_mode_admin = not other_user_exists

            if setup_mode_admin:
                # Try to add to admins group
                admin_result = cast(
                    CursorResult,
                    session.execute(
                        insert(UserGroups)
                        .values(
                            user_id=user_obj.user_id,
                            group_id=admins_group.group_id,
                        )
                        .on_conflict_do_nothing()
                    ),
                )

                if admin_result.rowcount > 0:
                    logger.info(
                        f"Added first user '{user.user_name}' to {DefaultGroup.ADMINS} "
                        "group",
                        prefix="Login",
                    )

            session.commit()

            return LoginResponse(
                user=User(
                    user_name=user_obj.name,
                    email=user_obj.email,
                ),
                setup_mode_admin=setup_mode_admin,
            )

    # Permissions management

    def check_permission(  # noqa: D102
        self,
        user_name: str,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> bool:
        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                return False

            # Get user's group IDs
            user_group_ids = session.scalars(
                select(UserGroups.group_id).where(UserGroups.user_id == user.user_id)
            ).all()

            if not user_group_ids:
                return False

            # Get permissions that would satisfy this check
            sufficient_permissions = PERMISSION_GRANTS[permission]

            # Check permissions based on resource type
            if resource == BackendResourceType.SYSTEM:
                # Check system permissions
                grant = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.is_system == True,  # noqa: E712
                            Permissions.group_id.in_(user_group_ids),
                            Permissions.permission.in_(sufficient_permissions),
                        )
                    )
                )
            else:
                # Check collection permissions
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                grant = session.scalar(
                    select(Permissions).where(
                        and_(
                            Permissions.collection_id == collection.collection_id,
                            Permissions.group_id.in_(user_group_ids),
                            Permissions.permission.in_(sufficient_permissions),
                        )
                    )
                )

            return grant is not None

    def get_permissions(  # noqa: D102
        self,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> list[PermissionGrant]:
        with MBDB.get_session() as session:
            if resource == BackendResourceType.SYSTEM:
                # Get system permissions
                permissions_query = (
                    select(Permissions)
                    .where(Permissions.is_system == True)  # noqa: E712
                    .join(Groups, Permissions.group_id == Groups.group_id)
                )
            else:
                # Get collection permissions
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                permissions_query = (
                    select(Permissions)
                    .where(Permissions.collection_id == collection.collection_id)
                    .join(Groups, Permissions.group_id == Groups.group_id)
                )

            permissions_orm = session.scalars(permissions_query).all()

            grants = [
                PermissionGrant(
                    group_name=perm.group.name,
                    permission=PermissionType(perm.permission),
                )
                for perm in permissions_orm
            ]

            return grants

    def grant_permission(  # noqa: D102
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        with MBDB.get_session() as session:
            grant_permission(
                session=session,
                group_name=group_name,
                permission=permission,
                resource=resource,
            )
            session.commit()

    def revoke_permission(  # noqa: D102
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        with MBDB.get_session() as session:
            # Get group
            group = session.scalar(select(Groups).where(Groups.name == group_name))
            if not group:
                raise MatchboxGroupNotFoundError(f"Group '{group_name}' not found")

            if resource == BackendResourceType.SYSTEM:
                # Revoke system permission
                result = session.execute(
                    delete(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.is_system == True,  # noqa: E712
                        )
                    )
                )
            else:
                # Revoke collection permission
                collection = session.scalar(
                    select(Collections).where(Collections.name == resource)
                )
                if not collection:
                    raise MatchboxCollectionNotFoundError(name=resource)

                result = session.execute(
                    delete(Permissions).where(
                        and_(
                            Permissions.group_id == group.group_id,
                            Permissions.permission == permission,
                            Permissions.collection_id == collection.collection_id,
                        )
                    )
                )

            session.commit()

            result = cast(CursorResult, result)

            if result.rowcount > 0:
                logger.info(
                    f"Revoked {permission} permission on '{resource}' "
                    f"from group '{group_name}'",
                    prefix="Revoke permission",
                )
            else:
                logger.info(
                    f"Permission {permission} on '{resource}' "
                    f"not present for group '{group_name}'",
                    prefix="Revoke permission",
                )

    # Data management

    def validate_ids(self, ids: list[int]) -> bool:  # noqa: D102
        with MBDB.get_session() as session:
            data_inner_join = (
                session.query(Clusters)
                .filter(
                    Clusters.cluster_id.in_(
                        bindparam(
                            "ins_ids",
                            ids,
                            expanding=True,
                        )
                    )
                )
                .all()
            )

        existing_ids = {item.cluster_id for item in data_inner_join}
        missing_ids = set(ids) - existing_ids

        if missing_ids:
            raise MatchboxDataNotFound(
                message="Some items don't exist in Clusters table.",
                table=Clusters.__tablename__,
                data=list(missing_ids),
            )

        return True

    def dump(self) -> MatchboxSnapshot:  # noqa: D102
        return dump()

    def drop(self, certain: bool) -> None:  # noqa: D102
        if certain:
            MBDB.drop_database()
            Groups.initialise()
        else:
            raise MatchboxDeletionNotConfirmed(
                "This operation will drop the entire database and recreate it."
                "It's not expected to be used as part normal operations."
                "If you're sure you want to continue, rerun with certain=True"
            )

    def clear(self, certain: bool) -> None:  # noqa: D102
        if certain:
            MBDB.clear_database()
            Groups.initialise()
        else:
            raise MatchboxDeletionNotConfirmed(
                "This operation will drop all rows in the database but not the "
                "tables themselves. It's primarily used to reset following tests."
                "If you're sure you want to continue, rerun with certain=True"
            )

    def restore(self, snapshot: MatchboxSnapshot) -> None:  # noqa: D102
        if snapshot.backend_type != MatchboxBackends.POSTGRES:
            raise TypeError(
                f"Cannot restore {snapshot.backend_type} snapshot to PostgreSQL backend"
            )

        MBDB.clear_database()

        restore(
            snapshot=snapshot,
            batch_size=self.settings.batch_size,
        )

    def delete_orphans(self) -> int:  # noqa: D102
        with MBDB.get_session() as session:
            # Get all cluster ids in related tables
            union_all_cte = union_all(
                select(EvalJudgements.endorsed_cluster_id.label("cluster_id")),
                select(EvalJudgements.shown_cluster_id.label("cluster_id")),
                select(ClusterSourceKey.cluster_id),
                select(ResolverClusters.cluster_id),
                select(ModelEdges.left_id.label("cluster_id")),
                select(ModelEdges.right_id.label("cluster_id")),
            ).cte("union_all_cte")

            # Deduplicate only once
            not_orphans = (
                select(union_all_cte.c.cluster_id).distinct().cte("not_orphans")
            )

            # Return clusters not in related tables
            stmt = delete(Clusters).where(
                ~select(not_orphans.c.cluster_id)
                .where(not_orphans.c.cluster_id == Clusters.cluster_id)
                .exists()
            )
            # Delete orphans
            deletion = cast(CursorResult, session.execute(stmt))

            session.commit()

            logger.info(f"Deleted {deletion.rowcount} orphans", prefix="Delete orphans")
            return deletion.rowcount



import pytest

from matchbox.common.dtos import (
    Collection,
    Match,
    StepPath,
)


def test_match_validates() -> None:
    """Match objects are validated when they're instantiated."""
    source_path = StepPath(name="source", collection="default", run=1)
    target_path = StepPath(name="target", collection="default", run=1)
    Match(
        cluster=1,
        source=source_path,
        source_id={"a"},
        target=target_path,
        target_id={"b"},
    )

    # Missing source_id with target_id
    with pytest.raises(ValueError):
        Match(cluster=1, source=source_path, target=target_path, target_id={"b"})

    # Missing cluster with target_id
    with pytest.raises(ValueError):
        Match(source=source_path, source_id={"a"}, target=target_path, target_id={"b"})

    # Missing source_id with cluster
    with pytest.raises(ValueError):
        Match(cluster=1, source=source_path, target=target_path)


def test_validate_collection() -> None:
    """Default run name needs to be within all runs."""
    # No default
    Collection(runs=[1, 2])

    # Valid default
    Collection(default_run=2, runs=[1, 2])

    with pytest.raises(ValueError):
        Collection(default_run=2, runs=[1])



"""Collection commands for Matchbox CLI."""

from typing import Annotated

import typer
from rich import print
from rich.table import Table

from matchbox.client import _handler
from matchbox.client.cli.annotations import (
    CollectionOpt,
    DeletionOpt,
    GroupOpt,
    PermissionOpt,
)
from matchbox.common.dtos import DefaultGroup, PermissionGrant, ResourceOperationStatus

app = typer.Typer(help="Manage collections")


@app.callback(invoke_without_command=True)
def list_collections(ctx: typer.Context) -> None:
    """List all collections."""
    if ctx.invoked_subcommand is not None:
        return

    collections = _handler.list_collections()
    for collection in collections:
        print(collection)


@app.command("create")
def create_collection(
    name: CollectionOpt,
    admin_group: Annotated[
        str,
        typer.Option(
            "--group",
            "-g",
            help="Group that will administer this collection",
        ),
    ] = DefaultGroup.PUBLIC,
) -> None:
    """Create a new collection."""
    response: ResourceOperationStatus = _handler.create_collection(
        name, admin_group=admin_group
    )
    if response.success:
        print(f"✓ Created collection {name}")
        if admin_group != DefaultGroup.PUBLIC:
            print(f"  Admin permission granted to group {admin_group}")
    else:
        print(f"✗ Failed to create collection {name}")
        if response.details:
            print(f"  {response.details}")
        raise typer.Exit(code=1)


@app.command("delete")
def delete_collection(
    name: CollectionOpt,
    certain: DeletionOpt,
) -> None:
    """Delete a collection."""
    response: ResourceOperationStatus = _handler.delete_collection(
        name, certain=certain
    )
    if response.success:
        print(f"✓ Deleted collection {name}")
    else:
        print(f"✗ Failed to delete collection {name}")
        if response.details:
            print(f"  {response.details}")
        raise typer.Exit(code=1)


# Collection permissions


@app.command("permissions")
def list_permissions(collection: CollectionOpt) -> None:
    """List permissions for a collection."""
    permissions: list[PermissionGrant] = _handler.get_collection_permissions(collection)

    table = Table(title=f"Permissions for {collection}")
    table.add_column("Group", style="cyan")
    table.add_column("Permission", style="green")

    for perm in permissions:
        table.add_row(perm.group_name, perm.permission)

    if not permissions:
        print("[dim]No permissions granted[/dim]")
    else:
        print(table)


@app.command("grant")
def grant_permission(
    collection: CollectionOpt,
    group: GroupOpt,
    permission: PermissionOpt,
) -> None:
    """Grant a permission to a group on a collection."""
    response: ResourceOperationStatus = _handler.grant_collection_permission(
        collection=collection,
        group_name=group,
        permission=permission,
    )
    if response.success:
        print(f"✓ Granted {permission} on {collection} to {group}")
    else:
        print(f"✗ Failed to grant {permission} on {collection} to {group}")
        if response.details:
            print(f"  {response.details}")
        raise typer.Exit(code=1)


@app.command("revoke")
def revoke_permission(
    collection: CollectionOpt,
    group: GroupOpt,
    permission: PermissionOpt,
) -> None:
    """Revoke a permission from a group on a collection."""
    response: ResourceOperationStatus = _handler.revoke_collection_permission(
        collection=collection,
        group_name=group,
        permission=permission,
    )
    if response.success:
        print(f"✓ Revoked {permission} on {collection} from {group}")
    else:
        print(f"✗ Failed to revoke {permission} on {collection} from {group}")
        if response.details:
            print(f"  {response.details}")
        raise typer.Exit(code=1)



"""Evaluation PostgreSQL mixin for Matchbox server."""

from datetime import UTC, datetime, timedelta
from itertools import chain
from typing import TYPE_CHECKING, Any

import numpy as np
import polars as pl
import pyarrow as pa
from pyarrow import Table
from sqlalchemy import BIGINT, func, select
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import Session

from matchbox.common.arrow import (
    SCHEMA_CLUSTER_EXPANSION,
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_JUDGEMENTS,
)
from matchbox.common.datatypes import require
from matchbox.common.db import QueryReturnType, sql_to_df
from matchbox.common.dtos import ResolverStepPath
from matchbox.common.eval import Judgement as CommonJudgement
from matchbox.common.exceptions import (
    MatchboxTooManySamplesRequested,
    MatchboxUserNotFoundError,
)
from matchbox.common.transform import hash_cluster_leaves
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    EvalJudgements,
    ResolverClusters,
    SourceConfigs,
    Steps,
    Users,
)
from matchbox.server.postgresql.utils.db import (
    compile_sql,
    ingest_to_temporary_table,
)

if TYPE_CHECKING:
    from pyarrow import Table as ArrowTable

    from matchbox.server.postgresql.adapter.admin import MatchboxPostgresAdminMixin

    # The evaluation mixin has self functions that reference the admin mixin
    # This lets us refernce it in a typesafe way
    _MixinBase = MatchboxPostgresAdminMixin
else:
    ArrowTable = Any
    _MixinBase = object


class MatchboxPostgresEvaluationMixin(_MixinBase):
    """Evaluation mixin for the PostgreSQL adapter for Matchbox."""

    def insert_judgement(self, user_name: str, judgement: CommonJudgement) -> None:  # noqa: D102
        def get_or_create_cluster(leaves: list[int], session: Session) -> int:
            # Compute hash corresponding to set of source clusters (leaves)
            leaf_hashes = [
                session.scalars(
                    select(Clusters.cluster_hash).where(Clusters.cluster_id == leaf_id)
                ).one()
                for leaf_id in leaves
            ]
            cluster_hash = hash_cluster_leaves(leaf_hashes)

            # Upsert cluster
            result = session.execute(
                insert(Clusters)
                .values(cluster_hash=cluster_hash)
                .on_conflict_do_nothing(index_elements=["cluster_hash"])
                .returning(Clusters.cluster_id)
            )

            # Get cluster_id
            newly_created = (cluster_id := result.scalar_one_or_none()) is not None

            if not newly_created:
                cluster_id = session.scalars(
                    select(Clusters.cluster_id).where(
                        Clusters.cluster_hash == cluster_hash
                    )
                ).one()

            # Only insert Contains relationships if we created a new cluster
            if newly_created:
                for leaf_id in leaves:
                    session.execute(
                        insert(Contains)
                        .values(root=cluster_id, leaf=leaf_id)
                        .on_conflict_do_nothing()
                    )

            return require(cluster_id)

        # Check that all referenced leaf IDs exist
        ids = list(chain(*judgement.endorsed)) + judgement.shown
        self.validate_ids(ids)

        with MBDB.get_session() as session:
            # Check that the user exists
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            # Get or create shown cluster
            shown_id = get_or_create_cluster(judgement.shown, session)

            # Insert all judgements
            for leaves in judgement.endorsed:
                endorsed_id = get_or_create_cluster(leaves, session)
                session.execute(
                    insert(EvalJudgements).values(
                        user_id=user.user_id,
                        tag=judgement.tag,
                        shown_cluster_id=shown_id,
                        endorsed_cluster_id=endorsed_id,
                        timestamp=datetime.now(UTC),
                    )
                )

            session.commit()

    def get_judgements(self, tag: str | None = None) -> tuple[Table, Table]:  # noqa: D102
        def _cast_tables(
            judgements: pl.DataFrame, cluster_expansion: pl.DataFrame
        ) -> tuple[pa.Table, pa.Table]:
            """Cast judgement tables to conform to data transfer schema."""
            judgements = judgements.cast(pl.Schema(SCHEMA_JUDGEMENTS))
            cluster_expansion = cluster_expansion.cast(
                pl.Schema(SCHEMA_CLUSTER_EXPANSION)
            )

            return (
                judgements.to_arrow().cast(SCHEMA_JUDGEMENTS),
                cluster_expansion.to_arrow().cast(SCHEMA_CLUSTER_EXPANSION),
            )

        judgements_stmt = select(
            Users.name.label("user_name"),
            EvalJudgements.endorsed_cluster_id.label("endorsed"),
            EvalJudgements.shown_cluster_id.label("shown"),
        ).join(Users)

        if tag:
            judgements_stmt = judgements_stmt.where(EvalJudgements.tag == tag)

        with MBDB.get_adbc_connection() as conn:
            judgements = sql_to_df(
                stmt=compile_sql(judgements_stmt),
                connection=conn,
                return_type=QueryReturnType.POLARS,
            )

        if not len(judgements):
            cluster_expansion = pl.DataFrame(schema=pl.Schema(SCHEMA_CLUSTER_EXPANSION))
            return _cast_tables(judgements, cluster_expansion)

        shown_clusters = set(judgements["shown"].to_list())
        endorsed_clusters = set(judgements["endorsed"].to_list())
        referenced_clusters = Table.from_pydict(
            {"root": list(shown_clusters | endorsed_clusters)}
        )

        with ingest_to_temporary_table(
            table_name="judgements",
            schema_name="mb",
            column_types={
                "root": BIGINT(),
            },
            data=referenced_clusters,
        ) as temp_table:
            cluster_expansion_stmt = (
                select(temp_table.c.root, func.array_agg(Contains.leaf).label("leaves"))
                .select_from(temp_table)
                .join(Contains, Contains.root == temp_table.c.root)
                .group_by(temp_table.c.root)
            )

            with MBDB.get_adbc_connection() as conn:
                cluster_expansion = sql_to_df(
                    stmt=compile_sql(cluster_expansion_stmt),
                    connection=conn,
                    return_type=QueryReturnType.POLARS,
                )

        return _cast_tables(judgements, cluster_expansion)

    def sample_for_eval(  # noqa: D102
        self, n: int, path: ResolverStepPath, user_name: str
    ) -> ArrowTable:
        """Sample some clusters from a resolver step."""
        # Not currently checking validity of the user_name
        # If the user ID does not exist, the exclusion by previous judgements breaks
        if n > 100:
            # This reasonable assumption means simple "IS IN" function later is fine
            raise MatchboxTooManySamplesRequested(
                "Can only sample 100 entries at a time."
            )

        with MBDB.get_session() as session:
            # Get user
            user = session.scalar(select(Users).where(Users.name == user_name))
            if not user:
                raise MatchboxUserNotFoundError(f"User '{user_name}' not found")

            # Use ORM to get step metadata
            resolver_step = Steps.from_path(path=path, session=session)
            step_id = resolver_step.step_id

        # Get cluster IDs and features for this resolver and user.
        user_judgements = (
            select(EvalJudgements)
            .where(EvalJudgements.user_id == user.user_id)
            .subquery()
        )
        cluster_features_stmt = (
            select(
                ResolverClusters.cluster_id,
                # Practically our client explicitly avoids showing the same cluster to
                # a user more than oce, so conflicting decisions over time shouldn't be
                # possible. However, the ORM doesn't enforce this, so we take only the
                # latest decision here
                func.max(user_judgements.c.timestamp).label("latest_ts"),
            )
            .join(
                user_judgements,
                ResolverClusters.cluster_id == user_judgements.c.shown_cluster_id,
                isouter=True,
            )
            .where(
                ResolverClusters.step_id == step_id,
            )
            .group_by(ResolverClusters.cluster_id)
        )

        with MBDB.get_adbc_connection() as conn:
            cluster_features = sql_to_df(
                stmt=compile_sql(cluster_features_stmt),
                connection=conn,
                return_type=QueryReturnType.POLARS,
            )

        # Exclude clusters recently judged by this user
        to_sample = cluster_features.filter(
            (pl.col("latest_ts") < datetime.now(UTC) - timedelta(days=365))
            | (pl.col("latest_ts").is_null())
        )

        # Return early if nothing to sample from
        if not len(to_sample):
            return pl.DataFrame(schema=pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()

        # With fewer clusters than requested, return all
        if to_sample.shape[0] <= n:
            sampled_cluster_ids = to_sample.select("cluster_id").to_series().to_list()
        else:
            indices = np.random.choice(to_sample.shape[0], size=n, replace=False)
            sampled_cluster_ids = (
                to_sample[indices].select("cluster_id").to_series().to_list()
            )

        # Get all info we need for the cluster IDs we've sampled, i.e.:
        # source cluster IDs, keys and source steps
        with MBDB.get_adbc_connection() as conn:
            source_clusters = (
                select(Contains.root, Contains.leaf)
                .where(Contains.root.in_(sampled_cluster_ids))
                .subquery()
            )

            # The same leaf can be reused to represent rows across different sources.
            # Only retrieve info for sources upstream of this resolver step.
            source_step_ids = [
                res.step_id for res in resolver_step.ancestors if res.type == "source"
            ]
            source_steps = (
                select(Steps.name, Steps.step_id)
                .where(Steps.step_id.in_(source_step_ids))
                .subquery()
            )

            enrich_stmt = (
                select(
                    source_clusters.c.root,
                    source_clusters.c.leaf,
                    ClusterSourceKey.key,
                    source_steps.c.name.label("source"),
                )
                .select_from(source_clusters)
                .join(
                    ClusterSourceKey,
                    ClusterSourceKey.cluster_id == source_clusters.c.leaf,
                )
                .join(
                    SourceConfigs,
                    SourceConfigs.source_config_id == ClusterSourceKey.source_config_id,
                )
                .join(
                    source_steps,
                    source_steps.c.step_id == SourceConfigs.step_id,
                )
            )

            final_samples = sql_to_df(
                stmt=compile_sql(enrich_stmt),
                connection=conn,
                return_type=QueryReturnType.POLARS,
            )
            return final_samples.cast(pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()
